"""
Parsing using TODO

Includes classes:

- LangNameParser

"""

import re
from typing import List
from lute.parse.base import ParsedToken, AbstractParser


# TODO fix names
class LangNameParser(AbstractParser):
    """
    A parser for LANGNAME
    """

    @classmethod
    def name(cls):
        return "Lute LangName"

    @classmethod
    def uses_data_directory(cls):
        "Uses the data_directory (defined in the AbstractParser)."
        return False  # or True

    # @classmethod
    # def init_data_directory(cls):
    #     "Set up necessary files."
    #     pass

    def get_parsed_tokens(self, text: str, language) -> List[ParsedToken]:
        """
        Returns ParsedToken array for given language.
        """

        # Ensure standard carriage returns so that paragraph
        # markers are used correctly.  Lute uses paragraph markers
        # for rendering.
        text = text.replace("\r\n", "\n")

        words = []  # ... get words using parser.
        tokens = []
        pattern = f"[{language.word_characters}]"
        for word in words:
            is_word_char = re.match(pattern, word) is not None
            is_end_of_sentence = word in language.regexp_split_sentences
            if word == "\n":
                word = "¶"
            if word == "¶":
                is_word_char = False
                is_end_of_sentence = True
            t = ParsedToken(word, is_word_char, is_end_of_sentence)
            tokens.append(t)
        return tokens

    def get_reading(self, text: str):
        """
        Get reading -- some parsers can return readings.
        """
        return None



"""
Reading helpers.
"""

import functools
from lute.models.term import Term, Status
from lute.models.book import Text
from lute.book.stats import Service as StatsService
from lute.read.render.service import Service as RenderService
from lute.read.render.calculate_textitems import get_string_indexes
from lute.term.model import Repository

# from lute.utils.debug_helpers import DebugTimer


class Service:
    "Service."

    def __init__(self, session):
        self.session = session

    def set_unknowns_to_known(self, text: Text):
        """
        Given a text, create new Terms with status Well-Known
        for any new Terms.
        """
        rs = RenderService(self.session)
        paragraphs = rs.get_paragraphs(text.text, text.book.language)
        self._save_new_status_0_terms(paragraphs)

        unknowns = [
            ti.term
            for para in paragraphs
            for sentence in para
            for ti in sentence
            if ti.is_word and ti.term.status == 0
        ]

        batch_size = 100
        i = 0

        for t in unknowns:
            t.status = Status.WELLKNOWN
            self.session.add(t)
            i += 1
            if i % batch_size == 0:
                self.session.commit()

        # Commit any remaining.
        self.session.commit()

    def bulk_status_update(self, text: Text, terms_text_array, new_status):
        """
        Given a text and list of terms, update or create new terms
        and set the status.
        """
        language = text.book.language
        repo = Repository(self.session)
        for term_text in terms_text_array:
            t = repo.find_or_new(language.id, term_text)
            t.status = new_status
            repo.add(t)
        repo.commit()

    def _save_new_status_0_terms(self, paragraphs):
        "Add status 0 terms for new textitems in paragraph."
        tis_with_new_terms = [
            ti
            for para in paragraphs
            for sentence in para
            for ti in sentence
            if ti.is_word and ti.term.id is None and ti.term.status == 0
        ]

        for ti in tis_with_new_terms:
            self.session.add(ti.term)
        self.session.commit()

    def start_reading(self, dbbook, pagenum):
        "Start reading a page in the book, getting paragraphs."

        text = dbbook.text_at_page(pagenum)
        text.load_sentences()

        svc = StatsService(self.session)
        svc.mark_stale(dbbook)
        dbbook.current_tx_id = text.id
        self.session.add(dbbook)
        self.session.add(text)
        self.session.commit()

        lang = text.book.language
        rs = RenderService(self.session)
        paragraphs = rs.get_paragraphs(text.text, lang)
        self._save_new_status_0_terms(paragraphs)

        return paragraphs

    def get_popup_data(self, termid):
        "Get popup data, or None if popup shouldn't be shown."
        term = self.session.get(Term, termid)

        def has_popup_data(cterm):
            return (
                (cterm.translation or "").strip() != ""
                or (cterm.romanization or "").strip() != ""
                or cterm.get_current_image() is not None
            )

        if not has_popup_data(term) and len(term.parents) == 0:
            return None

        term_tags = [tt.text for tt in term.term_tags]

        def make_array(t):
            ret = {
                "term": t.text,
                "roman": t.romanization,
                "trans": t.translation if t.translation else "-",
                "tags": [tt.text for tt in t.term_tags],
            }
            return ret

        parent_terms = [p.text for p in term.parents]
        parent_terms = ", ".join(parent_terms)

        parents = term.parents
        if len(parents) == 1 and parents[0].translation == term.translation:
            parents = []
        parent_data = [make_array(p) for p in parents]

        def sort_components(components):
            # Sort components by min position in string and length.
            component_and_pos = []
            for c in components:
                c_indices = [
                    loc[1] for loc in get_string_indexes([c.text_lc], term.text_lc)
                ]

                # Sometimes the components aren't found
                # in the string, which makes no sense ...
                # ref https://github.com/LuteOrg/lute-v3/issues/474
                if len(c_indices) > 0:
                    component_and_pos.append([c, min(c_indices)])

            def compare(a, b):
                # Lowest position (closest to front of string) sorts first.
                if a[1] != b[1]:
                    return -1 if (a[1] < b[1]) else 1
                # Longest sorts first.
                alen = len(a[0].text)
                blen = len(b[0].text)
                return -1 if (alen > blen) else 1

            component_and_pos.sort(key=functools.cmp_to_key(compare))
            return [c[0] for c in component_and_pos]

        rs = RenderService(self.session)
        components = [
            c
            for c in rs.find_all_Terms_in_string(term.text, term.language)
            if c.id != term.id
        ]
        components = sort_components(components)

        component_data = [make_array(c) for c in components]
        component_data = [c for c in component_data if c["trans"] != "-"]

        images = [term.get_current_image()] if term.get_current_image() else []
        for p in term.parents:
            if p.get_current_image():
                images.append(p.get_current_image())
        # DISABLED CODE: Don't include component images in the hover for now,
        # it can get confusing!
        # ref https://github.com/LuteOrg/lute-v3/issues/355
        # for c in components:
        #     if c.get_current_image():
        #         images.append(c.get_current_image())

        images = list(set(images))

        return {
            "term": term,
            "flashmsg": term.get_flash_message(),
            "term_tags": term_tags,
            "term_images": images,
            "parentdata": parent_data,
            "parentterms": parent_terms,
            "components": component_data,
        }



"""
Read service tests.
"""

from lute.models.term import Term
from lute.book.model import Book, Repository
from lute.read.service import Service
from lute.db import db

from tests.dbasserts import assert_record_count_equals, assert_sql_result


def test_set_unknowns_to_known(english, app_context):
    "Unknowns (status 0) or new are set to well known."
    t = Term(english, "dog")
    db.session.add(t)
    db.session.commit()

    b = Book()
    b.title = "blah"
    b.language_id = english.id
    b.text = "Dog CAT dog cat."
    r = Repository(db.session)
    dbbook = r.add(b)
    r.commit()

    sql = "select WoTextLC, WoStatus from words order by WoText"
    assert_sql_result(sql, ["dog; 1"], "before start")

    service = Service(db.session)
    service.start_reading(dbbook, 1)
    assert_sql_result(sql, ["cat; 0", "dog; 1"], "after start")

    tx = dbbook.texts[0]
    tx.text = "Dog CAT dog cat extra."
    db.session.add(tx)
    db.session.commit()

    service = Service(db.session)
    service.set_unknowns_to_known(tx)
    assert_sql_result(sql, ["cat; 99", "dog; 1", "extra; 99"], "after set")


def test_smoke_start_reading(english, app_context):
    "Smoke test book."
    b = Book()
    b.title = "blah"
    b.language_id = english.id
    b.text = "Here is some content.  Here is more."
    r = Repository(db.session)
    dbbook = r.add(b)
    r.commit()

    assert_record_count_equals("select * from sentences", 0, "before start")
    service = Service(db.session)
    service.start_reading(dbbook, 1)
    assert_record_count_equals("select * from sentences", 2, "after start")


def test_start_reading_creates_Terms_for_unknown_words(english, app_context):
    "Unknown (status 0) terms are created for all new words."
    t = Term(english, "dog")
    db.session.add(t)
    db.session.commit()

    b = Book()
    b.title = "blah"
    b.language_id = english.id
    b.text = "Dog CAT dog cat."
    r = Repository(db.session)
    dbbook = r.add(b)
    r.commit()

    sql = "select WoTextLC from words order by WoText"
    assert_sql_result(sql, ["dog"], "before start")

    service = Service(db.session)
    paragraphs = service.start_reading(dbbook, 1)
    textitems = [
        ti
        for para in paragraphs
        for sentence in para
        for ti in sentence
        if ti.is_word and ti.wo_id is None
    ]
    assert (
        len(textitems) == 0
    ), f"All text items should have a term, but got {textitems}"
    assert_sql_result(sql, ["cat", "dog"], "after start")



"""
Forms while reading.
"""

from flask_wtf import FlaskForm
from wtforms import TextAreaField


class TextForm(FlaskForm):
    """
    Text page - for editing a page.
    """

    text = TextAreaField("Text")



"""
Term parent map file generation tests.
"""

import os
import tempfile
import pytest
from lute.db import db
from lute.term_parent_map.service import Service
from tests.utils import add_terms, make_book


@pytest.fixture(name="output_tempfile")
def fixture_temp_file():
    "Yield a temp file."
    with tempfile.NamedTemporaryFile() as t:
        f = t.name
    yield f
    os.unlink(f)


@pytest.fixture(name="_book")
def fixture_terms_and_book(spanish):
    "Create a book."
    content = "Hola tengo un gato.  No tengo una lista.\nElla tiene una bebida."
    book = make_book("Hola", content, spanish)
    yield book


def assert_file_content(fname, expected):
    "Assert file name fname contains expected."
    expected.sort()
    with open(fname, "r", encoding="utf-8") as f:
        actual = f.read().splitlines()
        actual.sort()
        assert expected == actual, "contents"


def test_smoke_book_file_created(app_context, _book, output_tempfile):
    "Smoke test only."
    service = Service(db.session)
    service.export_unknown_terms(_book, output_tempfile)
    expected = [
        "hola",
        "gato",
        "lista",
        "tengo",
        "un",
        "no",
        "una",
        "ella",
        "tiene",
        "bebida",
    ]
    assert_file_content(output_tempfile, expected)


def test_known_multiword_terms_can_hide_unknown_terms(
    spanish, app_context, _book, output_tempfile
):
    "See what terms would show up as blue 'unknown' for a book."
    spanish_terms = ["gato", "lista", "tiene una", "listo"]
    add_terms(spanish, spanish_terms)
    service = Service(db.session)
    service.export_unknown_terms(_book, output_tempfile)
    expected = ["hola", "tengo", "un", "no", "ella", "tiene", "una", "bebida"]
    assert_file_content(output_tempfile, expected)


# def test_book_file_contains_status_0_words(app_context, _book, output_tempfile):
#    "Status 0 words are created as 'placeholders', and should be included."



"""
ThaiParser tests.
"""


import pytest

# pylint: disable=wrong-import-order
from lute.models.term import Term
from lute.parse.base import ParsedToken

from lute_thai_parser.parser import ThaiParser


def test_token_count(thai):
    """
    token_count checks.
    """
    cases = [
        ("สวัสดี", 1),
        ("ลาก่อน", 1),
        ("ฉันรักคุณ", 3),
        ("ฉันกำลังเรียนภาษาไทย", 4),
    ]
    for text, expected_count in cases:
        t = Term(thai, text)
        assert t.token_count == expected_count, text
        assert t.text_lc == t.text, "case"


def assert_tokens_equals(text, lang, expected):
    """
    Parsing a text using a language should give the expected parsed tokens.

    expected is given as array of:
    [ original_text, is_word, is_end_of_sentence ]
    """
    p = ThaiParser()
    actual = p.get_parsed_tokens(text, lang)
    expected = [ParsedToken(*a) for a in expected]
    assert [str(a) for a in actual] == [str(e) for e in expected]


def test_end_of_sentence_stored_in_parsed_tokens(thai):
    """
    ParsedToken is marked as EOS=True at ends of sentences.
    """
    s = "สวัสดีทุกคน! ฉันเรียนภาษาไทยมา2เดือนแล้วฯ"

    expected = [
        ("สวัสดี", True),
        ("ทุกคน", True),
        ("!", False, True),
        (" ", False, True),
        ("ฉัน", True),
        ("เรียน", True),
        ("ภาษาไทย", True),
        ("มา", True),
        ("2", False),
        ("เดือน", True),
        ("แล้ว", True, False),
        ("ฯ", False, True),
    ]
    assert_tokens_equals(s, thai, expected)


def test_carriage_returns_treated_as_reverse_p_character(thai):
    """
    Returns need to be marked with the backwards P for rendering etc.
    """
    s = "สวัสดีทุกคน!\nฉันเรียนภาษาไทยมา2เดือนแล้ว"

    expected = [
        ("สวัสดี", True),
        ("ทุกคน", True),
        ("!", False, True),
        ("¶", False, True),
        ("ฉัน", True),
        ("เรียน", True),
        ("ภาษาไทย", True),
        ("มา", True),
        ("2", False),
        ("เดือน", True),
        ("แล้ว", True, False),
    ]
    assert_tokens_equals(s, thai, expected)



"""
/book routes.
"""

import json
from flask import (
    Blueprint,
    request,
    jsonify,
    render_template,
    redirect,
    flash,
)
from lute.utils.data_tables import DataTablesFlaskParamParser
from lute.book.service import Service, BookImportException
from lute.book.datatables import get_data_tables_list
from lute.book.forms import NewBookForm, EditBookForm
from lute.book.stats import Service as StatsService
import lute.utils.formutils
from lute.db import db

from lute.models.language import Language
from lute.models.repositories import (
    BookRepository,
    UserSettingRepository,
    LanguageRepository,
)
from lute.book.model import Book, Repository


bp = Blueprint("book", __name__, url_prefix="/book")


def _load_term_custom_filters(request_form, parameters):
    "Manually add filters that the DataTablesFlaskParamParser doesn't know about."
    filter_param_names = [
        "filtLanguage",
    ]
    request_params = request_form.to_dict(flat=True)
    for p in filter_param_names:
        parameters[p] = request_params.get(p)


def datatables_source(is_archived):
    "Get datatables json for books."
    # In the future, we might want to create an API such as
    # get_books(sort_order, search_string, length, index, language_id).
    # See DataTablesFlaskParamParser.parse_params_2(request.form)
    # (currently unused)
    parameters = DataTablesFlaskParamParser.parse_params(request.form)
    _load_term_custom_filters(request.form, parameters)
    data = get_data_tables_list(parameters, is_archived, db.session)
    return jsonify(data)


@bp.route("/datatables/active", methods=["POST"])
def datatables_active_source():
    "Datatables data for active books."
    return datatables_source(False)


@bp.route("/archived", methods=["GET"])
def archived():
    "List archived books."
    language_choices = lute.utils.formutils.language_choices(
        db.session, "(all languages)"
    )
    current_language_id = lute.utils.formutils.valid_current_language_id(db.session)

    return render_template(
        "book/index.html",
        status="Archived",
        language_choices=language_choices,
        current_language_id=current_language_id,
    )


# Archived must be capitalized, or the ajax call 404's.
@bp.route("/datatables/Archived", methods=["POST"])
def datatables_archived_source():
    "Datatables data for archived books."
    return datatables_source(True)


def _book_from_url(url):
    "Create a new book, or flash an error if can't parse."
    b = Book()
    service = Service()
    try:
        b = service.book_from_url(url)
    except BookImportException as e:
        flash(e.message, "notice")
        b = Book()
    return b


def _language_is_rtl_map():
    """
    Return language-id to is_rtl map, to be used during book creation.
    """
    ret = {}
    for lang in db.session.query(Language).all():
        ret[lang.id] = lang.right_to_left
    return ret


@bp.route("/new", methods=["GET", "POST"])
def new():
    "Create a new book, either from text or from a file."
    b = Book()
    import_url = request.args.get("importurl", "").strip()
    if import_url != "":
        b = _book_from_url(import_url)

    form = NewBookForm(obj=b)
    form.language_id.choices = lute.utils.formutils.language_choices(db.session)
    repo = Repository(db.session)

    if form.validate_on_submit():
        try:
            form.populate_obj(b)
            book = repo.add(b)
            repo.commit()
            return redirect(f"/read/{book.id}/page/1", 302)
        except BookImportException as e:
            flash(e.message, "notice")

    # Don't set the current language before submit.
    usrepo = UserSettingRepository(db.session)
    current_language_id = int(usrepo.get_value("current_language_id"))
    form.language_id.data = current_language_id

    return render_template(
        "book/create_new.html",
        book=b,
        form=form,
        tags=repo.get_book_tags(),
        rtl_map=json.dumps(_language_is_rtl_map()),
        show_language_selector=True,
    )


@bp.route("/edit/<int:bookid>", methods=["GET", "POST"])
def edit(bookid):
    "Edit a book - can only change a few fields."
    repo = Repository(db.session)
    b = repo.load(bookid)
    form = EditBookForm(obj=b)

    if form.validate_on_submit():
        form.populate_obj(b)
        repo.add(b)
        repo.commit()
        flash(f"{b.title} updated.")
        return redirect("/", 302)

    lang_repo = LanguageRepository(db.session)
    lang = lang_repo.find(b.language_id)
    return render_template(
        "book/edit.html",
        book=b,
        title_direction="rtl" if lang.right_to_left else "ltr",
        form=form,
        tags=repo.get_book_tags(),
    )


@bp.route("/import_webpage", methods=["GET", "POST"])
def import_webpage():
    return render_template("book/import_webpage.html")


def _find_book(bookid):
    "Find book from db."
    br = BookRepository(db.session)
    return br.find(bookid)


@bp.route("/archive/<int:bookid>", methods=["POST"])
def archive(bookid):
    "Archive a book."
    b = _find_book(bookid)
    b.archived = True
    db.session.add(b)
    db.session.commit()
    return redirect("/", 302)


@bp.route("/unarchive/<int:bookid>", methods=["POST"])
def unarchive(bookid):
    "Archive a book."
    b = _find_book(bookid)
    b.archived = False
    db.session.add(b)
    db.session.commit()
    return redirect("/", 302)


@bp.route("/delete/<int:bookid>", methods=["POST"])
def delete(bookid):
    "Archive a book."
    b = _find_book(bookid)
    db.session.delete(b)
    db.session.commit()
    return redirect("/", 302)


@bp.route("/table_stats/<int:bookid>", methods=["GET"])
def table_stats(bookid):
    "Get the stats, return ajax."
    b = _find_book(bookid)
    if b is None or b.language is None:
        # Playwright tests were sometimes passing an id that didn't exist ...
        # I believe this is due to page caching, i.e. the book listing
        # is showing books and IDs that no longer exist after cache reset.
        # TODO fix_hack: get rid of this hack.
        return jsonify({})
    svc = StatsService(db.session)
    stats = svc.get_stats(b)
    ret = {
        "distinctterms": stats.distinctterms,
        "distinctunknowns": stats.distinctunknowns,
        "unknownpercent": stats.unknownpercent,
        "status_distribution": stats.status_distribution,
    }
    return jsonify(ret)



"""
Term popup data tests.
"""

import pytest
from lute.models.term import Term, Status
from lute.read.service import Service
from lute.db import db


@pytest.fixture(name="service")
def fixture_svc(app_context):
    "Service"
    return Service(db.session)


def test_popup_data_is_none_if_no_data(spanish, app_context, service):
    "Return None if no popup."
    t = Term(spanish, "gato")
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d is None, "No data, no popup"

    t.translation = "hello"
    d = service.get_popup_data(t.id)
    assert d is not None, "Have data, popup"


def test_popup_shown_if_parent_exists_even_if_no_other_data(
    spanish, app_context, service
):
    "I always want to know if a parent has been set."
    t = Term(spanish, "gato")
    db.session.add(t)
    db.session.commit()
    d = service.get_popup_data(t.id)
    assert d is None, "No data, no popup"

    p = Term(spanish, "perro")
    t.parents.append(p)
    db.session.add(t)
    db.session.commit()
    d = service.get_popup_data(t.id)
    assert d is not None, "Has parent, popup, even if no other data."


def test_popup_data_is_shown_if_have_data_regardless_of_status(
    spanish, app_context, service
):
    "Return None if no-popup statuses."
    t = Term(spanish, "gato")
    t.translation = "hello"
    for s in [1, Status.UNKNOWN, Status.IGNORED]:
        t.status = s
        db.session.add(t)
        db.session.commit()
        d = service.get_popup_data(t.id)
        assert d is not None, f"Have data for status {s}"


def test_term_with_no_parents(spanish, app_context, service):
    "Keep the lights on test, smoke only."
    t = Term(spanish, "gato")
    t.translation = "cat"
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d["parentdata"] == [], "no parents"
    assert d["parentterms"] == "", "no parents"


def test_single_parent_data_not_added_if_same_translation(
    spanish, app_context, service
):
    "Extra data added for display."
    t = Term(spanish, "gato")
    t.translation = "cat"
    p = Term(spanish, "perro")
    p.translation = "cat"
    t.parents.append(p)
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d["parentdata"] == [], "no extra parent data"
    assert d["parentterms"] == "perro", "one parent"


def test_parent_data_added_if_parent_translation_is_different(
    spanish, app_context, service
):
    "Extra data added for display."
    t = Term(spanish, "gato")
    t.translation = "cat"
    p = Term(spanish, "perro")
    p.translation = "kitty"
    t.parents.append(p)
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    expected_p_data = {
        "term": "perro",
        "roman": None,
        "trans": "kitty",
        "tags": [],
    }
    assert d["parentdata"] == [expected_p_data], "extra parent data added"
    assert d["parentterms"] == "perro", "one parent"


def test_parent_data_always_added_if_multiple_parents(spanish, app_context, service):
    "Extra data added for display."
    t = Term(spanish, "gato")
    t.translation = "cat"
    p = Term(spanish, "perro")
    p.translation = "kitty"
    p2 = Term(spanish, "hombre")
    p2.translation = "kitteh"
    t.parents.append(p)
    t.parents.append(p2)
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    expected_p1_data = {"term": "perro", "roman": None, "trans": "kitty", "tags": []}
    expected_p2_data = {"term": "hombre", "roman": None, "trans": "kitteh", "tags": []}
    expected = [expected_p1_data, expected_p2_data]
    assert d["parentdata"] == expected, "extra parent data added"
    assert d["parentterms"] == "perro, hombre", "parents"


## Component term checks.


def make_terms(term_trans_pairs, spanish):
    "Make test data"
    for c in term_trans_pairs:
        ct = Term(spanish, c[0])
        ct.translation = c[1]
        db.session.add(ct)


def assert_components(d, expected, msg=""):
    "Check components."

    def _c_to_string(c):
        s = "; ".join([c["term"], c["trans"]])
        zws = chr(0x200B)
        return s.replace(zws, "")

    actual = [_c_to_string(c) for c in d["components"]]
    assert actual == expected, msg


def test_single_term_not_included_in_own_components(spanish, app_context, service):
    "Keep the lights on test, smoke only."
    t = Term(spanish, "gato")
    t.translation = "cat"
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d["components"] == [], "no components"


def test_component_without_translation_not_returned(spanish, app_context, service):
    "Component word is returned."
    t = Term(spanish, "un gato")
    t.translation = "a cat"
    db.session.add(t)
    make_terms([("gato", "")], spanish)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d["components"] == [], "no data"


def test_component_word_with_translation_returned(spanish, app_context, service):
    "Component word is returned."
    t = Term(spanish, "un gato")
    t.translation = "a cat"
    db.session.add(t)
    make_terms([("gato", "cat")], spanish)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert_components(d, ["gato; cat"], "one component")


def test_nested_multiword_components(spanish, app_context, service):
    "Complete components are returned."
    t = Term(spanish, "un gato gordo")
    t.translation = "a fat cat"
    db.session.add(t)
    make_terms([("gato", "cat"), ("gat", "x"), ("un gato", "a cat")], spanish)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert_components(d, ["un gato; a cat", "gato; cat"], "components")


def test_multiword_components_returned_in_order_of_appearance(
    spanish, app_context, service
):
    "Complete components are returned."
    t = Term(spanish, "un gato gordo")
    t.translation = "a fat cat"
    db.session.add(t)
    make_terms(
        [
            ("gato", "cat"),
            ("gat", "x"),
            ("gato gordo", "fat"),
            ("un gato", "a cat"),
        ],
        spanish,
    )
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert_components(
        d, ["un gato; a cat", "gato gordo; fat", "gato; cat"], "components"
    )


def test_components_only_returned_once(spanish, app_context, service):
    "Component not returned multiple times if present multiple times."
    t = Term(spanish, "un gato gordo gato")
    t.translation = "a cat fat cat"
    db.session.add(t)
    make_terms([("gato", "cat")], spanish)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert_components(d, ["gato; cat"], "components")



"""
Book tests.
"""

import textwrap
from lute.models.book import Book
from lute.db import db
from tests.dbasserts import assert_sql_result


# pylint: disable=too-many-arguments,too-many-positional-arguments
def assert_add(book, pagenum, before, text, expected, msg=""):
    "Assert that adding results in expected table content."
    sql = f"""
    select TxText, TxOrder from Texts
    where TxBkID = {book.id}
    order by TxOrder"""

    t = None
    if before:
        t = book.add_page_before(pagenum)
    else:
        t = book.add_page_after(pagenum)
    t.text = text
    db.session.add(book)
    db.session.commit()
    assert_sql_result(sql, expected, msg)


def assert_remove(book, pagenum, expected, msg):
    "assert removed"
    sql = f"""
    select TxText, TxOrder from Texts
    where TxBkID = {book.id}
    order by TxOrder"""

    book.remove_page(pagenum)

    db.session.add(book)
    db.session.commit()
    assert_sql_result(sql, expected, msg)


def test_can_add_page(app_context, english):
    "Add page before and after."
    fulltext = textwrap.dedent(
        """\
      1
      ---
      2
      ---
      3
    """
    )
    b = Book.create_book("hi", english, fulltext)
    db.session.add(b)
    db.session.commit()

    sql = f"""
    select TxText, TxOrder from Texts
    where TxBkID = {b.id}
    order by TxOrder"""
    assert_sql_result(sql, ["1; 1", "2; 2", "3; 3"], "initial book")

    assert_add(b, 2, True, "B2", ["1; 1", "B2; 2", "2; 3", "3; 4"], "new")
    assert_add(b, 3, False, "A4", ["1; 1", "B2; 2", "2; 3", "A4; 4", "3; 5"], "new")


def test_add_page_before_first(app_context, english):
    "Add page before and after."
    b = Book.create_book("hi", english, "hi")
    db.session.add(b)
    db.session.commit()

    assert_add(b, 1, True, "B1", ["B1; 1", "hi; 2"], "new")
    assert_add(b, 0, True, "B0", ["B0; 1", "B1; 2", "hi; 3"], "0")
    assert_add(b, -100, True, "B-", ["B-; 1", "B0; 2", "B1; 3", "hi; 4"], "neg")
    assert_add(b, 100, True, "B+", ["B-; 1", "B0; 2", "B1; 3", "B+; 4", "hi; 5"], "big")


def test_add_page_after_last(app_context, english):
    "Add page after last."
    b = Book.create_book("hi", english, "hi")
    db.session.add(b)
    db.session.commit()

    assert_add(b, 1, False, "B1", ["hi; 1", "B1; 2"], "new")
    assert_add(b, 100, False, "B100", ["hi; 1", "B1; 2", "B100; 3"], "100")
    assert_add(b, -100, False, "Bneg", ["hi; 1", "Bneg; 2", "B1; 3", "B100; 4"], "neg")


def test_can_remove_page(app_context, english):
    "Remove pages before and after."
    fulltext = textwrap.dedent(
        """\
      1
      ---
      2
      ---
      3
    """
    )
    b = Book.create_book("hi", english, fulltext)
    db.session.add(b)
    db.session.commit()

    assert_remove(b, 2, ["1; 1", "3; 2"], "2nd page removed")
    assert_remove(b, 12, ["1; 1", "3; 2"], "bad page removal ignored")
    assert_remove(b, -12, ["1; 1", "3; 2"], "bad page removal ignored")
    assert_remove(b, 0, ["1; 1", "3; 2"], "bad page removal ignored")
    assert_remove(b, 1, ["3; 1"], "1st removed")
    assert_remove(b, 1, ["3; 1"], "can't remove sole page")



"""
Parsing using khmer-nltk

Includes classes:

- KhmerParser

"""

import re

from typing import List

import khmernltk

from lute.parse.base import ParsedToken, AbstractParser


class KhmerParser(AbstractParser):
    """
    A parser for KHMER
    """

    @classmethod
    def name(cls):
        return "Lute Khmer"

    @classmethod
    def uses_data_directory(cls):
        "Uses the data_directory (defined in the AbstractParser)."
        return False  # or True

    # @classmethod
    # def init_data_directory(cls):
    #     "Set up necessary files."
    #     pass

    def _handle_special_token(self, token: str, special_char: str) -> List[str]:
        """
        Handle special token scenarios by replacing all special tokens with newline characters.

        Example:
            If \ is the special token then
            "\hey\man\\\" will evaluate as
            ["\n", "hey", "\n", "man", "\n", "\n", "\n"]
        """
        if token == special_char:
            return ["\n"]

        num_leading_slashes = len(token) - len(token.lstrip(special_char))
        num_trailing_slashes = len(token) - len(token.rstrip(special_char))
        output = []

        output.extend("\n" * num_leading_slashes)

        tokens = token.strip(special_char).split(special_char)

        if len(tokens) == 1:
            output.append(tokens[0])
        else:
            for token in tokens[:-1]:
                output.append(token)
                output.append("\n")
            output.append(tokens[-1])

        output.extend("\n" * num_trailing_slashes)

        return output

    def word_tokenize(self, text: str) -> List[str]:
        """
        Tokenize a text using khmernltk and handle the fact that khmernltk
        completely omits newline characters by replacing all newline chars with
        something that khmernltk won't omit.
        """
        special_char = "\\"
        text = text.replace("\n", special_char)
        output = []

        for token in khmernltk.word_tokenize(text):
            if special_char in token:
                output.extend(self._handle_special_token(token, special_char))
                continue
            output.append(token)
        return output

    def get_parsed_tokens(self, text: str, language) -> List[ParsedToken]:
        """
        Returns ParsedToken array for given language.
        """

        # Ensure standard carriage returns so that paragraph
        # markers are used correctly.  Lute uses paragraph markers
        # for rendering.
        text = text.replace("\r\n", "\n")
        words = self.word_tokenize(text)  # ... get words using parser.
        pattern = f"[{language.word_characters}]"
        tokens = []
        for word in words:
            is_end_of_sentence = word in language.regexp_split_sentences
            is_word_char = (not is_end_of_sentence) and (
                re.match(pattern, word) is not None
            )
            if word == "\n":
                word = "¶"
            if word == "¶":
                is_word_char = False
                is_end_of_sentence = True

            t = ParsedToken(word, is_word_char, is_end_of_sentence)
            tokens.append(t)
        return tokens

    def get_reading(self, text: str):
        """
        Get reading -- some parsers can return readings.
        """
        return None



"""
get_string_indexes tests.
"""

import pytest
from lute.read.render.multiword_indexer import MultiwordTermIndexer

zws = "\u200B"  # zero-width space


@pytest.mark.parametrize(
    "name,terms,tokens,expected",
    [
        ("empty", [], ["a"], []),
        ("no terms", [], ["a"], []),
        ("no tokens", ["a"], [], []),
        ("no match", ["x"], ["a"], []),
        ("single match", ["a"], ["a"], [("a", 0)]),
        ("single match 2", ["a"], ["b", "a"], [("a", 1)]),
        ("same term twice", ["a"], ["b", "a", "c", "a"], [("a", 1), ("a", 3)]),
        (
            "multiple terms",
            ["a", "b"],
            ["b", "a", "c", "a"],
            [("b", 0), ("a", 1), ("a", 3)],
        ),
        ("multi-word term", [f"a{zws}b"], ["b", "a", "b", "a"], [(f"a{zws}b", 1)]),
        (
            "repeated m-word term",
            [f"a{zws}a"],
            ["a", "a", "a", "b"],
            [(f"a{zws}a", 0), (f"a{zws}a", 1)],
        ),
        ("bound check term at end", ["a"], ["b", "c", "a"], [("a", 2)]),
    ],
)
def test_scenario(name, terms, tokens, expected):
    "Test scenario."
    mw = MultiwordTermIndexer()
    for t in terms:
        mw.add(t)
    results = list(mw.search_all(tokens))
    assert len(results) == len(expected), name
    assert results == expected, name


def test_can_search_multiple_times_with_different_tokens():
    "Single match, returns token index."
    mw = MultiwordTermIndexer()
    mw.add("a")
    results = list(mw.search_all(["a", "b"]))
    assert len(results) == 1, "one match"
    assert results[0] == ("a", 0)

    results = list(mw.search_all(["b", "a"]))
    assert len(results) == 1, "one match"
    assert results[0] == ("a", 1)



"""
Development-only routes.

These routes are accessed during acceptance tests.
The acceptance tests don't have easy access to
the db model ... at least, as far as I can tell ...
so this is a quick-and-easy way to set some test
data state.

If there's a better way to manage this, I'll change
this.

These are dangerous methods that hack on the db data without
safeguards, so they can only be run against a test_ db.
"""

import os
from sqlalchemy import text
from flask import Blueprint, current_app, Response, jsonify, redirect, flash
from lute.models.language import Language
from lute.models.repositories import UserSettingRepository
import lute.parse.registry
from lute.db import db
import lute.db.management
from lute.db.demo import Service as DemoService


bp = Blueprint("dev_api", __name__, url_prefix="/dev_api")


@bp.before_request
def _ensure_is_test_db():
    "These methods should only be run on a test db!"
    if not current_app.env_config.is_test_db:
        raise RuntimeError("Dangerous method, only possible on test_lute database.")


@bp.route("/wipe_db", methods=["GET"])
def wipe_db():
    "Clean it all."
    lute.db.management.delete_all_data(db.session)
    flash("db wiped")
    return redirect("/", 302)


@bp.route("/load_demo", methods=["GET"])
def load_demo():
    "Clean out everything, and load the demo."
    lute.db.management.delete_all_data(db.session)
    demosvc = DemoService(db.session)
    demosvc.set_load_demo_flag()
    demosvc.load_demo_data()
    flash("demo loaded")
    return redirect("/", 302)


@bp.route("/load_demo_languages", methods=["GET"])
def load_demo_languages():
    "Clean out everything, and load the demo langs with dummy dictionaries."
    lute.db.management.delete_all_data(db.session)
    demosvc = DemoService(db.session)
    demosvc.load_demo_languages()
    langs = db.session.query(Language).all()
    for lang in langs:
        d = lang.dictionaries[0]
        d.dicturi = f"/dev_api/dummy_dict/{lang.name}/[LUTE]"
        d.dicttype = "embeddedhtml"  # Ensure not pop-up
        db.session.add(lang)
    db.session.commit()
    return redirect("/", 302)


@bp.route("/load_demo_stories", methods=["GET"])
def load_demo_stories():
    "Stories only.  No db wipe."
    demosvc = DemoService(db.session)
    demosvc.load_demo_stories()
    flash("stories loaded")
    return redirect("/", 302)


@bp.route("/language_ids", methods=["GET"])
def get_lang_ids():
    "Get ids of all langs."
    langs = db.session.query(Language).all()
    ret = {}
    for lang in langs:
        ret[lang.name] = lang.id
    return jsonify(ret)


@bp.route("/delete_all_terms", methods=["GET"])
def delete_all_terms():
    "Delete all the terms only."
    db.session.query(text("DELETE FROM words"))
    db.session.commit()
    flash("terms deleted")
    return redirect("/", 302)


@bp.route("/sqlresult/<string:sql>", methods=["GET"])
def sql_result(sql):
    "Get the sql result as json."

    def clean_val(v):
        if v is None:
            return "NULL"
        if isinstance(v, str) and "\u200b" in v:
            return v.replace("\u200b", "/")
        return v

    content = []
    result = db.session.execute(text(sql)).all()
    for row in result:
        rowvals = [clean_val(v) for v in row]
        content.append("; ".join(map(str, rowvals)))

    return jsonify(content)


@bp.route("/dummy_dict/<string:langname>/<string:term>", methods=["GET"])
def dummy_language_dict(langname, term):
    "Fake language dictionary/term lookup."
    return Response(f"dev_api/dummy_dict/{langname}/{term}")


@bp.route("/disable_parser/<string:parsername>/<string:renameto>", methods=["GET"])
def disable_parser(parsername, renameto):
    "Hack: rename a parser in the registry so that languages can't find it."
    p = lute.parse.registry.__LUTE_PARSERS__
    if parsername in p:
        p[renameto] = p.pop(parsername)
    langs = db.session.query(Language).all()
    unsupported = [
        {"parser_type": lang.parser_type, "language": lang.name}
        for lang in langs
        if not lang.is_supported
    ]
    allparsers = [
        {"parser_type": lang.parser_type, "language": lang.name} for lang in langs
    ]
    available_types = list(p.keys())
    return jsonify(
        {
            "_unsupported": unsupported,
            "available_types": available_types,
            "parsers": allparsers,
        }
    )


@bp.route("/disable_backup", methods=["GET"])
def disable_backup():
    "Disables backup -- tests don't need to back up."
    repo = UserSettingRepository(db.session)
    repo.set_value("backup_enabled", False)
    db.session.commit()
    flash("backup disabled")
    return redirect("/", 302)


@bp.route("/throw_error/<message>", methods=["GET"])
def throw_error(message):
    "Throw an error to ensure handler works!"
    raise RuntimeError(message)


@bp.route("/fake_story.html", methods=["GET"])
def fake_story():
    "Return a fake story for import book test."
    return Response(
        """<html>
    <head>
    <title>Mi perro.</title>
    </head>
    <body>
    <p>Hola. Tengo un perro.</p>
    </body>
    </html>"""
    )


@bp.route("/temp_file_content/<filename>", methods=["GET"])
def temp_file_content(filename):
    "Get the content of the file."
    fpath = os.path.join(current_app.env_config.temppath, filename)
    s = ""
    with open(fpath, "r", encoding="utf-8") as f:
        s = f.read()
    return s



"""
Export language definition and first two pages of all books to lute/db/language_defs/

Run this as a module from the root directory, with languages to export as args:

python -m scripts.dump_lang_data Catalan Bulgarian

NOTE: this script assumes that the config is in the root folder, or at lute/config/config.yml.
"""

import sys
import os
import yaml

from lute.models.book import Book
from lute.models.language import Language
from lute.db import db
import lute.app_factory


def _language_defs_path():
    "Path to the definitions and stories."
    thisdir = os.path.dirname(__file__)
    d = os.path.join(thisdir, "..", "lute", "db", "language_defs")
    return os.path.abspath(d)


def _partial_book_content(b):
    "Get book content as string."
    fulltext = [t.text for t in b.texts]
    first_two_pages = fulltext[:2]
    s = "\n".join(first_two_pages).replace("\r", "").strip()
    return f"# title: {b.title}\n\n{s}"


def _write_langs(language_names, outdir):
    "Write all langs and books."
    langs = db.session.query(Language).all()
    langs = [lang for lang in langs if lang.name.lower() in language_names]
    for lang in langs:
        ld = lang.to_dict()
        n = ld["name"].lower().replace(" ", "_")
        langdir = os.path.join(outdir, n)
        if not os.path.exists(langdir):
            os.mkdir(langdir)

        file_path = os.path.join(langdir, "definition.yaml")
        with open(file_path, "w", encoding="utf-8") as fp:
            yaml.dump(ld, stream=fp, allow_unicode=True, sort_keys=False)
        print(lang.name)

        books = db.session.query(Book).filter(Book.language == lang).all()
        books = [b for b in books if not b.archived]
        story_count = 1
        for b in books:
            filename = f"story_{story_count}"
            story_count += 1
            file_path = os.path.join(langdir, f"{filename}.txt")
            with open(file_path, "w", encoding="utf-8") as f:
                f.write(_partial_book_content(b))
            file_size_kb = os.path.getsize(file_path) / 1024
            print(f"- {filename} ({file_size_kb:.2f} KB)")


def main(langnames):
    "Entry point."
    outputdir = _language_defs_path()
    print(f"Outputting to {outputdir}")

    langnames = [n.lower() for n in langnames]
    langnames = list(set(langnames))
    app = lute.app_factory.create_app()
    with app.app_context():
        _write_langs(langnames, outputdir)


#####################
# Entry point.

if len(sys.argv) < 2:
    print("Language names required.")
    sys.exit(0)

main(sys.argv[1:])



"""
Bulk import books.
"""

import csv
import os
import sys

from lute.book.model import Book, Repository
from lute.db import db
from lute.models.repositories import LanguageRepository


def import_books_from_csv(file, language, tags, commit):
    """
    Bulk import books from a CSV file.

    Args:

      file:     the path to the CSV file to import (see lute/cli/commands.py
                for the requirements for this file).
      language: the name of the language to use by default, as it appears in
                your languages settings
      tags:     a list of tags to apply to all books
      commit:   a boolean value indicating whether to commit the changes to the
                database. If false, a list of books to be imported will be
                printed out, but no changes will be made.
    """
    repo = Repository(db.session)
    lang_repo = LanguageRepository(db.session)

    count = 0
    with open(file, newline="", encoding="utf-8") as f:
        r = csv.DictReader(f)
        for row in r:
            book = Book()
            book.title = row["title"]
            book.language_name = row.get("language") or language
            if not book.language_name:
                print(f"Skipping book with unspecified language: {book.title}")
                continue
            lang = lang_repo.find_by_name(book.language_name)
            if not lang:
                print(
                    f"Skipping book with unknown language ({book.language_name}): {book.title}"
                )
                continue
            if repo.find_by_title(book.title, lang.id) is not None:
                print(f"Already exists in {book.language_name}: {book.title}")
                continue
            count += 1
            all_tags = []
            if tags:
                all_tags.extend(tags)
            if "tags" in row and row["tags"]:
                for tag in row["tags"].split(","):
                    if tag and tag not in all_tags:
                        all_tags.append(tag)
            book.book_tags = all_tags
            book.text = row["text"]
            book.source_uri = row.get("url") or None
            if "audio" in row and row["audio"]:
                book.audio_filename = os.path.join(os.path.dirname(file), row["audio"])
            book.audio_bookmarks = row.get("bookmarks") or None
            repo.add(book)
            print(
                f"Added {book.language_name} book (tags={','.join(all_tags)}): {book.title}"
            )

    print()
    print(f"Added {count} books")
    print()

    if not commit:
        db.session.rollback()
        print("Dry run, no changes made.")
        return

    print("Committing...")
    sys.stdout.flush()
    repo.commit()



"""
KhmerParser tests.
"""


import pytest

# pylint: disable=wrong-import-order
from lute.models.term import Term
from lute.parse.base import ParsedToken

from lute_khmer_parser.parser import KhmerParser


def test_dummy_test():
    "A dummy test so that pytest doesn't complain in github ci."
    s = "Hello"
    assert s == "Hello", "TODO - fix these tests for your parser :-)"


def test_token_count(khmer):
    """
    token_count checks.
    """
    cases = [
        ("ជំរាបសួរ", 2),
        ("ខ្ញុំ", 1),
        ("ខ្ញុំស្រលាញ់អ្នក។", 4),
        ("ខ្ញុំរៀនភាសាខ្មែរ", 4),
        ("ខ្ញុំចូលចិត្តរៀនភាសាខ្មែរជាមួយមិត្តរបស់ខ្ញុំ", 9),
    ]

    for text, expected_count in cases:
        t = Term(khmer, text)
        assert t.token_count == expected_count, text
        assert t.text_lc == t.text, "case"


def assert_tokens_equals(text, lang, expected):
    """
    Parsing a text using a language should give the expected parsed tokens.

    expected is given as array of:
    [ original_text, is_word, is_end_of_sentence ]
    """
    p = KhmerParser()
    actual = p.get_parsed_tokens(text, lang)
    expected = [ParsedToken(*a) for a in expected]
    assert [str(a) for a in actual] == [str(e) for e in expected]


def test_end_of_sentence_stored_in_parsed_tokens(khmer):
    """
    ParsedToken is marked as EOS=True at ends of sentences.
    """
    s = "ខ្ញុំចូលចិត្តរៀនភាសាខ្មែរជាមួយមិត្តរបស់ខ្ញុំ។ ចុះអ្នកវិញ?"

    expected = [
        ("ខ្ញុំ", True),
        ("ចូលចិត្ត", True),
        ("រៀន", True),
        ("ភាសា", True),
        ("ខ្មែរ", True),
        ("ជាមួយ", True),
        ("មិត្ត", True),
        ("របស់", True),
        ("ខ្ញុំ", True),
        ("។", False, True),
        (" ", False),
        ("ចុះ", True),
        ("អ្នក", True),
        ("វិញ", True),
        ("?", False, True),
    ]
    assert_tokens_equals(s, khmer, expected)


def test_carriage_returns_treated_as_reverse_p_character(khmer):
    """
    Returns need to be marked with the backwards P for rendering etc.
    """
    s = "ខ្ញុំចូលចិត្តរៀនភាសាខ្មែរជាមួយមិត្តរបស់ខ្ញុំ។\nចុះអ្នកវិញ?"

    expected = [
        ("ខ្ញុំ", True),
        ("ចូលចិត្ត", True),
        ("រៀន", True),
        ("ភាសា", True),
        ("ខ្មែរ", True),
        ("ជាមួយ", True),
        ("មិត្ត", True),
        ("របស់", True),
        ("ខ្ញុំ", True),
        ("។", False, True),
        ("¶", False, True),
        ("ចុះ", True),
        ("អ្នក", True),
        ("វិញ", True),
        ("?", False, True),
    ]
    assert_tokens_equals(s, khmer, expected)



"""
SpaceDelimitedParser tests.
"""

import sys
import re
import unicodedata
from lute.parse.space_delimited_parser import SpaceDelimitedParser
from lute.parse.base import ParsedToken


def assert_tokens_equals(text, lang, expected):
    """
    Parsing a text using a language should give the expected parsed tokens.

    expected is given as array of:
    [ original_text, is_word, is_end_of_sentence ]
    """
    p = SpaceDelimitedParser()
    actual = p.get_parsed_tokens(text, lang)
    expected = [ParsedToken(*a) for a in expected]
    assert [str(a) for a in actual] == [str(e) for e in expected]


def assert_string_equals(text, lang, expected):
    """
    Parsing a text with a language's settings should yield tokens.

    Similar to assert_tokens_equals, but it stringizes the tokens, e.g.:
    'Hi.\nGoodbye.' => '[Hi].¶[Goodbye].'
    """
    p = SpaceDelimitedParser()
    actual = p.get_parsed_tokens(text, lang)

    def to_string(tokens):
        ret = ""
        for tok in tokens:
            s = tok.token
            if tok.is_word:
                s = "[" + s + "]"
            ret += s
        return ret

    assert to_string(actual) == expected


def test_end_of_sentence_stored_in_parsed_tokens(spanish):
    "ParsedToken is marked as EOS=True at ends of sentences."
    s = "Tengo un gato.\nTengo dos."

    expected = [
        ("Tengo", True, False),
        (" ", False, False),
        ("un", True, False),
        (" ", False, False),
        ("gato", True, False),
        (".", False, True),
        ("¶", False, True),
        ("Tengo", True, False),
        (" ", False, False),
        ("dos", True, False),
        (".", False, True),
    ]

    assert_tokens_equals(s, spanish, expected)


def test_exceptions_are_considered_when_splitting_sentences(english):
    "Languages can have exceptions (like 'Mrs.') that shouldn't split sentences."
    s = "1. Mrs. Jones is here."

    expected = [
        ("1. ", False, True),
        ("Mrs.", True, False),
        (" ", False, False),
        ("Jones", True, False),
        (" ", False, False),
        ("is", True, False),
        (" ", False, False),
        ("here", True, False),
        (".", False, True),
    ]

    assert_tokens_equals(s, english, expected)


def test_single_que(spanish):
    """
    Sanity check: que with accent was getting mishandled.
    """
    text = "Tengo que y qué."
    expected = [
        ("Tengo", True, False),
        (" ", False, False),
        ("que", True, False),
        (" ", False, False),
        ("y", True, False),
        (" ", False, False),
        ("qué", True, False),
        (".", False, True),
    ]
    assert_tokens_equals(text, spanish, expected)


def test_EE_UU_exception_should_be_considered(spanish):
    """
    An exception containing multiple dots should be one single token.
    """
    s = "Estamos en EE.UU. hola."
    spanish.exceptions_split_sentences = "EE.UU."

    expected = [
        ("Estamos", True, False),
        (" ", False, False),
        ("en", True, False),
        (" ", False, False),
        ("EE.UU.", True, False),
        (" ", False, False),
        ("hola", True, False),
        (".", False, True),
    ]

    assert_tokens_equals(s, spanish, expected)


def test_just_EE_UU(spanish):
    """
    A sentence of a single word, where that word is an exception, should be a single token.
    """
    s = "EE.UU."
    spanish.exceptions_split_sentences = "EE.UU."
    expected = [
        ("EE.UU.", True, False),
    ]
    assert_tokens_equals(s, spanish, expected)


def test_quick_checks(english):
    "Fast sanity checks."
    assert_string_equals("test", english, "[test]")
    assert_string_equals("test.", english, "[test].")
    assert_string_equals('"test."', english, '"[test]."')
    assert_string_equals('"test".', english, '"[test]".')
    assert_string_equals("Hi there.", english, "[Hi] [there].")
    assert_string_equals("Hi there.  Goodbye.", english, "[Hi] [there]. [Goodbye].")
    assert_string_equals("Hi.\nGoodbye.", english, "[Hi].¶[Goodbye].")
    assert_string_equals("He123llo.", english, "[He]123[llo].")
    assert_string_equals("1234", english, "1234")
    assert_string_equals("1234.", english, "1234.")
    assert_string_equals("1234.Hello", english, "1234.[Hello]")


def test_zero_width_non_joiner_retained(german):
    """
    Verify zero-width joiner characters are retained in languages that
    include them as word characters.

    Test case from https://en.wikipedia.org/wiki/Zero-width_non-joiner.
    """
    assert_string_equals("Brotzeit", german, "[Brotzeit]")
    assert_string_equals("Brot\u200czeit", german, "[Brot\u200czeit]")


def test_zero_width_joiner_retained(hindi):
    """
    Verify zero-width joiner characters are retained in languages that
    include them as word characters.

    We see them used to replace Hindi conjunct characters with individual consonants.
    """
    assert_string_equals("namaste", hindi, "[namaste]")
    assert_string_equals("नमस्ते", hindi, "[नमस्ते]")
    assert_string_equals("नमस\u200dते", hindi, "[नमस\u200dते]")


def test_default_word_pattern_latin(generic):
    """
    Verify the default word pattern handles Latin alphabets (0000..00FF).
    """

    # Source: https://www.folger.edu/explore/shakespeares-works/henry-v/read/5/2/
    assert_string_equals(
        "Saint Denis be my speed!—donc vôtre est France, et vous êtes mienne.",
        generic,
        " ".join(
            [
                "[Saint]",
                "[Denis]",
                "[be]",
                "[my]",
                "[speed]!—[donc]",
                "[vôtre]",
                "[est]",
                "[France],",
                "[et]",
                "[vous]",
                "[êtes]",
                "[mienne].",
            ]
        ),
    )


def test_default_word_pattern_devanagari(generic):
    """
    Verify the default word pattern handles the Devanagari Unicode block (0900..097F).
    """

    # Source: https://en.wikipedia.org/wiki/Hindi#Sample_text
    assert_string_equals(
        "अनुच्छेद १(एक): सभी मनुष्य जन्म से स्वतन्त्र तथा मर्यादा और अधिकारों में समान होते हैं।",
        generic,
        " ".join(
            [
                "[अनुच्छेद]",
                "१([एक]):",
                "[सभी]",
                "[मनुष्य]",
                "[जन्म]",
                "[से]",
                "[स्वतन्त्र]",
                "[तथा]",
                "[मर्यादा]",
                "[और]",
                "[अधिकारों]",
                "[में]",
                "[समान]",
                "[होते]",
                "[हैं]।",
            ]
        ),
    )


def test_default_word_pattern_georgian(generic):
    """
    Verify the default word pattern handles the Georgian Unicode block (10A0..10FF).
    """

    # Source: https://en.wikipedia.org/wiki/Georgian_language#Examples
    assert_string_equals(
        "ყველა ადამიანი იბადება თავისუფალი და თანასწორი თავისი ღირსებითა და უფლებებით.",
        generic,
        " ".join(
            [
                "[ყველა]",
                "[ადამიანი]",
                "[იბადება]",
                "[თავისუფალი]",
                "[და]",
                "[თანასწორი]",
                "[თავისი]",
                "[ღირსებითა]",
                "[და]",
                "[უფლებებით].",
            ]
        ),
    )


def test_default_word_pattern_gothic(generic):
    """
    Verify the default word pattern handles the Gothic Unicode block (10330..1034F).
    This is an import test case because it tests larger Unicode values.
    """

    # Source: https://en.wikipedia.org/wiki/Gothic_language#Examples
    assert_string_equals(
        "𐌰𐍄𐍄𐌰 𐌿𐌽𐍃𐌰𐍂 𐌸𐌿 𐌹𐌽 𐌷𐌹𐌼𐌹𐌽𐌰𐌼", generic, "[𐌰𐍄𐍄𐌰] [𐌿𐌽𐍃𐌰𐍂] [𐌸𐌿] [𐌹𐌽] [𐌷𐌹𐌼𐌹𐌽𐌰𐌼]"
    )


def test_all_chars_in_categories_match_default_word_chars():
    "Default_word_chars builds a range of characters ... ensure chars in categories are all found."
    categories = set(["Cf", "Ll", "Lm", "Lo", "Lt", "Lu", "Mc", "Mn", "Sk"])

    word_chars = SpaceDelimitedParser.get_default_word_characters()
    pattern = rf"[{word_chars}]"
    regex = re.compile(pattern, flags=re.IGNORECASE)
    for i in range(1, sys.maxunicode):
        c = chr(i)
        if unicodedata.category(c) in categories:
            assert regex.match(c), f"Match for {c}"



"""
Repositories.
"""

from sqlalchemy import text as sqltext, and_, func
from lute.db import db
from lute.models.setting import UserSetting, BackupSettings, SystemSetting
from lute.models.language import Language
from lute.models.term import Term, TermTag
from lute.models.book import Book, BookTag


class SettingRepositoryBase:
    "Repository."

    def __init__(self, session, classtype):
        self.session = session
        self.classtype = classtype

    def key_exists_precheck(self, keyname):
        """
        Check key validity for certain actions.
        """

    def set_value(self, keyname, keyvalue):
        "Set, but don't save, a setting."
        self.key_exists_precheck(keyname)
        s = (
            self.session.query(self.classtype)
            .filter(self.classtype.key == keyname)
            .first()
        )
        if s is None:
            s = self.classtype()
            s.key = keyname
        s.value = keyvalue
        self.session.add(s)

    def key_exists(self, keyname):
        "True if exists."
        s = (
            self.session.query(self.classtype)
            .filter(self.classtype.key == keyname)
            .first()
        )
        no_key = s is None
        return not no_key

    def get_value(self, keyname):
        "Get the saved key, or None if it doesn't exist."
        self.key_exists_precheck(keyname)
        s = (
            self.session.query(self.classtype)
            .filter(self.classtype.key == keyname)
            .first()
        )
        if s is None:
            return None
        return s.value

    def delete_key(self, keyname):
        "Delete a key."
        s = (
            self.session.query(self.classtype)
            .filter(self.classtype.key == keyname)
            .first()
        )
        if s is not None:
            self.session.delete(s)


class MissingUserSettingKeyException(Exception):
    """
    Cannot set or get unknown user keys.
    """


class UserSettingRepository(SettingRepositoryBase):
    "Repository."

    def __init__(self, session):
        super().__init__(session, UserSetting)

    def key_exists_precheck(self, keyname):
        """
        User keys must exist.
        """
        if not self.key_exists(keyname):
            raise MissingUserSettingKeyException(keyname)

    def get_backup_settings(self):
        "Convenience method."
        bs = BackupSettings()

        def _bool(v):
            return v in (1, "1", "y", True)

        bs.backup_enabled = _bool(self.get_value("backup_enabled"))
        bs.backup_auto = _bool(self.get_value("backup_auto"))
        bs.backup_warn = _bool(self.get_value("backup_warn"))
        bs.backup_dir = self.get_value("backup_dir")
        bs.backup_count = int(self.get_value("backup_count") or 5)
        bs.last_backup_datetime = self.get_last_backup_datetime()
        return bs

    def get_last_backup_datetime(self):
        "Get the last_backup_datetime as int, or None."
        v = self.get_value("lastbackup")
        if v is None:
            return None
        return int(v)

    def set_last_backup_datetime(self, v):
        "Set and save the last backup time."
        self.set_value("lastbackup", v)
        self.session.commit()


class SystemSettingRepository(SettingRepositoryBase):
    "Repository."

    def __init__(self, session):
        super().__init__(session, SystemSetting)


class LanguageRepository:
    "Repository."

    def __init__(self, session):
        self.session = session

    def find(self, language_id):
        "Get by ID."
        return self.session.query(Language).filter(Language.id == language_id).first()

    def find_by_name(self, name):
        "Get by name."
        return (
            self.session.query(Language)
            .filter(func.lower(Language.name) == func.lower(name))
            .first()
        )

    def delete(self, language):
        """
        Hacky method to delete language and all terms, books, and dicts
        associated with it.

        There is _certainly_ a better way to do this using
        Sqlalchemy relationships and cascade deletes, but I
        was running into problems with it (things not cascading,
        or warnings ("SAWarning: Object of type <Term> not in
        session, add operation along 'Language.terms' will not
        proceed") during test runs.  It would be nice to have
        a "correct" mapping, but this is good enough for now.

        TODO zzfuture fix: fix Language-Book and -Term mappings.
        """
        sqls = [
            "pragma foreign_keys = ON",
            f"delete from languages where LgID = {language.id}",
        ]
        for s in sqls:
            self.session.execute(sqltext(s))
        self.session.commit()

    def all_dictionaries(self):
        "All dictionaries for all languages."
        lang_dicts = {}
        for lang in db.session.query(Language).all():
            lang_dicts[lang.id] = {
                "term": lang.active_dict_uris("terms"),
                "sentence": lang.active_dict_uris("sentences"),
            }
        return lang_dicts


class TermTagRepository:
    "Repository."

    def __init__(self, session):
        self.session = session

    def find(self, termtag_id):
        "Get by ID."
        return self.session.query(TermTag).filter(TermTag.id == termtag_id).first()

    def find_by_text(self, text):
        "Find a tag by text, or None if not found."
        return self.session.query(TermTag).filter(TermTag.text == text).first()

    def find_or_create_by_text(self, text):
        "Return tag or create one."
        ret = self.find_by_text(text)
        if ret is not None:
            return ret
        return TermTag(text)


class TermRepository:
    "Repository."

    def __init__(self, session):
        self.session = session

    def find_by_spec(self, spec):
        """
        Find by the given spec term's language ID and text.
        Returns None if not found.
        """
        langid = spec.language.id
        text_lc = spec.text_lc
        query = self.session.query(Term).filter(
            and_(Term.language_id == langid, Term.text_lc == text_lc)
        )
        terms = query.all()
        if not terms:
            return None
        return terms[0]

    def delete_empty_images(self):
        """
        Data clean-up: delete empty images.

        The code was leaving empty images in the db, which are obviously no good.
        This is a hack to clean up the data.
        """
        sql = "delete from wordimages where trim(WiSource) = ''"
        self.session.execute(sqltext(sql))
        self.session.commit()


class BookTagRepository:
    "Repository."

    def __init__(self, session):
        self.session = session

    def find_or_create_by_text(self, text):
        "Return tag or create one."
        ret = db.session.query(BookTag).filter(BookTag.text == text).first()
        if ret is not None:
            return ret
        return BookTag.make_book_tag(text)


class BookRepository:
    "Repository."

    def __init__(self, session):
        self.session = session

    def find(self, book_id):
        "Get by ID."
        return self.session.query(Book).filter(Book.id == book_id).first()

    def find_by_title(self, book_title, language_id):
        "Get by title."
        return (
            self.session.query(Book)
            .filter(and_(Book.title == book_title, Book.language_id == language_id))
            .first()
        )



"""
Book statistics.
"""

import json
from sqlalchemy import select, text
from lute.read.render.service import Service as RenderService
from lute.models.book import Book, BookStats
from lute.models.repositories import UserSettingRepository

# from lute.utils.debug_helpers import DebugTimer


class Service:
    "Service."

    def __init__(self, session):
        self.session = session

    def _last_n_pages(self, book, txindex, n):
        "Get next n pages, or at least n pages."
        start_index = max(0, txindex - n)
        end_index = txindex + n
        texts = book.texts[start_index:end_index]
        return texts[-n:]

    def _get_sample_texts(self, book):
        "Get texts to use as sample."
        txindex = 0
        if (book.current_tx_id or 0) != 0:
            for t in book.texts:
                if t.id == book.current_tx_id:
                    break
                txindex += 1

        repo = UserSettingRepository(self.session)
        sample_size = int(repo.get_value("stats_calc_sample_size") or 5)
        texts = self._last_n_pages(book, txindex, sample_size)
        return texts

    def calc_status_distribution(self, book):
        """
        Calculate statuses and count of unique words per status.

        Does a full render of a small number of pages
        to calculate the distribution.
        """

        # DebugTimer.clear_total_summary()
        # dt = DebugTimer("get_status_distribution", display=False)
        texts = self._get_sample_texts(book)

        # Getting the individual paragraphs per page, and then combining,
        # is much faster than combining all pages into one giant page.
        service = RenderService(self.session)
        mw = service.get_multiword_indexer(book.language)
        textitems = []
        for tx in texts:
            textitems.extend(service.get_textitems(tx.text, book.language, mw))
        # # Old slower code:
        # text_sample = "\n".join([t.text for t in texts])
        # paras = get_paragraphs(text_sample, book.language) ... etc.
        # dt.step("get_paragraphs")

        textitems = [ti for ti in textitems if ti.is_word]
        statterms = {0: [], 1: [], 2: [], 3: [], 4: [], 5: [], 98: [], 99: []}
        for ti in textitems:
            statterms[ti.wo_status or 0].append(ti.text_lc)

        stats = {}
        for statusval, allterms in statterms.items():
            uniques = list(set(allterms))
            statterms[statusval] = uniques
            stats[statusval] = len(uniques)

        # dt.step("compiled")
        # DebugTimer.total_summary()

        return stats

    def refresh_stats(self):
        "Refresh stats for all books requiring update."
        sql = "delete from bookstats where status_distribution is null"
        self.session.execute(text(sql))
        self.session.commit()
        book_ids_with_stats = select(BookStats.BkID).scalar_subquery()
        books_to_update = (
            self.session.query(Book).filter(~Book.id.in_(book_ids_with_stats)).all()
        )
        books = [b for b in books_to_update if b.is_supported]
        for book in books:
            stats = self._calculate_stats(book)
            self._update_stats(book, stats)

    def mark_stale(self, book):
        "Mark a book's stats as stale to force refresh."
        bk_id = book.id
        self.session.query(BookStats).filter_by(BkID=bk_id).delete()
        self.session.commit()

    def get_stats(self, book):
        "Gets stats from the cache if available, or calculates."
        bk_id = book.id
        stats = self.session.query(BookStats).filter_by(BkID=bk_id).first()
        if stats is None or stats.status_distribution is None:
            newstats = self._calculate_stats(book)
            self._update_stats(book, newstats)
            stats = self.session.query(BookStats).filter_by(BkID=bk_id).first()
        return stats

    def _calculate_stats(self, book):
        "Calc stats for the book using the status distribution."
        status_distribution = self.calc_status_distribution(book)
        unknowns = status_distribution[0]
        allunique = sum(status_distribution.values())

        percent = 0
        if allunique > 0:  # In case not parsed.
            percent = round(100.0 * unknowns / allunique)

        return {
            "allunique": allunique,
            "unknowns": unknowns,
            "percent": percent,
            "distribution": json.dumps(status_distribution),
        }

    def _update_stats(self, book, stats):
        "Update BookStats for the given book."
        s = self.session.query(BookStats).filter_by(BkID=book.id).first()
        if s is None:
            s = BookStats(BkID=book.id)
        s.distinctterms = stats["allunique"]
        s.distinctunknowns = stats["unknowns"]
        s.unknownpercent = stats["percent"]
        s.status_distribution = stats["distribution"]
        self.session.add(s)
        self.session.commit()



"""
Generate a rough data file for a given language for all terms.

Gets *all* books for a given language, and writes a data file to
the specified csv output file name.

Generates csv with headings:
term; count; familycount; books; definition; status; children

e.g.
term; count; familycount; books; definition; status; children
haber; 100; 1500; book1,book2; to exist; 99; hay (500), he (200), has (150) ...
"""

import csv
from lute.db import db
from lute.models.book import Book
from lute.read.render.service import Service


def _add_term_to_dict(t, terms):
    "Add term to dictionary and return it."
    key = t.text_lc
    if key in terms:
        return terms[key]

    tag_list = ", ".join([tg.text for tg in t.term_tags])
    if tag_list == "":
        tag_list = "-"

    parents_text = sorted([p.text_lc for p in t.parents])
    parents_text = "; ".join(parents_text)
    if parents_text == "":
        parents_text = "-"

    zws = "\u200B"
    hsh = {
        "sourceterm": t,
        "term": t.text.replace(zws, ""),
        "count": 0,
        "familycount": 0,
        "books": [],
        "definition": t.translation or "-",
        "status": t.status,
        "parents": parents_text,
        "children": [],
        "tags": tag_list,
    }
    terms[key] = hsh
    return hsh


def _process_book(b, terms, multiword_indexer):
    "Process pages in book, add to output."
    print(f"Processing {b.title} ...")
    i = 0
    service = Service(db.session)
    for text in b.texts:
        i += 1
        if i % 10 == 0:
            print(f"  page {i} of {b.page_count}", end="\r")
        textitems = service.get_textitems(text.text, b.language, multiword_indexer)
        displayed_terms = [
            ti.term for ti in textitems if ti.is_word and ti.term is not None
        ]
        for t in displayed_terms:
            e = _add_term_to_dict(t, terms)
            e["count"] += 1
            e["familycount"] += 1
            if b.title not in e["books"]:
                e["books"].append(b.title)

            for parent in t.parents:
                p = _add_term_to_dict(parent, terms)
                p["familycount"] += 1
                if b.title not in p["books"]:
                    p["books"].append(b.title)
                if t.text_lc not in p["children"]:
                    p["children"].append(t.text_lc)


def _book_list_truncated(title_array):
    "Return first 5 books, + count of rest."
    titles = list(set(title_array))
    first_5 = titles[:5]
    ret = ", ".join(first_5)
    count_rest = len(titles) - len(first_5)
    if count_rest > 0:
        ret += f" [... +{count_rest} more]"
    return ret


def _finalize_output(terms):
    "Convert terms hash to usable output."
    for _, hsh in terms.items():
        hsh["books"] = _book_list_truncated(hsh["books"])

        # children to child (count)
        children = []
        for key in hsh["children"]:
            t = terms[key]
            children.append({"count": t["count"], "term": t["sourceterm"].text})
        csorted = sorted(children, key=lambda c: c["count"], reverse=True)
        children_string = "; ".join([f"{c['term']} ({c['count']})" for c in csorted])
        if children_string == "":
            children_string = "-"
        hsh["children"] = children_string

    ret = terms.values()
    return sorted(ret, key=lambda x: (-x["familycount"], x["term"]))


def _load_indexers(books):
    "Load multiword indexers for book languages."
    service = Service(db.session)
    ret = {}
    lang_map = {book.language.id: book.language for book in books}
    for langid, lang in lang_map.items():
        ret[langid] = service.get_multiword_indexer(lang)
    return ret


def _generate_file(books, outfile_name):
    "Write data file for books to outfile_name."
    indexers = _load_indexers(books)
    terms = {}
    for b in books:
        _process_book(b, terms, indexers[b.language.id])
    outdata = _finalize_output(terms)

    with open(outfile_name, "w", newline="", encoding="utf-8") as outfile:
        keys = [
            "term",
            "count",
            "familycount",
            "books",
            "definition",
            "status",
            "parents",
            "children",
            "tags",
        ]
        writer = csv.DictWriter(outfile, fieldnames=keys, extrasaction="ignore")
        writer.writeheader()
        for r in outdata:
            writer.writerow(r)


def generate_language_file(language_name, outfile_name):
    """
    Generate the datafile for the language.
    """
    books = db.session.query(Book).all()
    books = [b for b in books if b.language.name == language_name]
    if len(books) == 0:
        print(f"No books for given language {language_name}, quitting.")
    else:
        print(f"Writing to {outfile_name}")
        _generate_file(books, outfile_name)
        print("Done.                     ")  # extra space overwrites old output.


def generate_book_file(bookid, outfile_name):
    """
    Generate the datafile for the book.
    """
    books = db.session.query(Book).all()
    books = [b for b in books if f"{b.id}" == f"{bookid}"]
    if len(books) == 0:
        print(f"No book with id = {bookid}.")
    else:
        print(f"Writing to {outfile_name}")
        _generate_file(books, outfile_name)
        print("Done.                     ")  # extra space overwrites old output.



"""
DB Backup tests.
"""

import os
from datetime import datetime, timezone
from unittest.mock import Mock, patch
import pytest

from lute.backup.service import Service, BackupException, DatabaseBackupFile
from lute.models.repositories import UserSettingRepository
from lute.db import db
from lute.language.service import Service as LanguageService

from tests.dbasserts import assert_record_count_equals

# pylint: disable=missing-function-docstring
# Test method names are pretty descriptive already.

utc = timezone.utc


@pytest.fixture(name="bkp_dir")
def fixture_backup_directory(testconfig):
    "Create clean backup dir."
    base_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../"))
    backup_dir = os.path.join(base_dir, "zz_bkp")
    if not os.path.exists(backup_dir):
        os.mkdir(backup_dir)

    # Cleanup prior to test.
    cleanup_directory(backup_dir)
    cleanup_directory(testconfig.userimagespath)

    yield backup_dir

    # Don't clean up after the test, in case want to check manually.
    # cleanup_directories(backup_dir, image_dir)


def cleanup_directory(directory):
    "Delete all files and subfolders."
    for root, dirs, files in os.walk(directory, topdown=False):
        for f in files:
            os.remove(os.path.join(root, f))
        for d in dirs:
            os.rmdir(os.path.join(root, d))


@pytest.fixture(name="backup_settings")
def fixture_backup_settings(app_context, bkp_dir):
    # app_context is passed so that the db session is available.
    repo = UserSettingRepository(db.session)
    ret = repo.get_backup_settings()
    ret.backup_dir = bkp_dir
    ret.backup_enabled = True
    yield ret


def _mock_backup_file(directory, name, time, size):
    file_mock = Mock()
    file_mock.name = name
    file_mock.path = os.path.join(directory, name)
    file_mock.size = size
    with open(file_mock.path, "wb") as f:
        f.seek(file_mock.size - 1)
        f.write(b"\0")
    os.utime(file_mock.path, (time.timestamp(), time.timestamp()))
    return file_mock


@pytest.fixture(name="auto_backup_file")
def fixture_auto_backup_file(bkp_dir):
    yield _mock_backup_file(
        bkp_dir,
        "lute_backup_2024-01-01-000000.db.gz",
        datetime(2024, 1, 1, 0, 0, 0, tzinfo=utc),
        567890,
    )


@pytest.fixture(name="manual_backup_file")
def fixture_manual_backup_file(bkp_dir):
    yield _mock_backup_file(
        bkp_dir,
        "manual_lute_backup_2024-02-01-000000.db.gz",
        datetime(2024, 2, 1, 0, 0, 0, tzinfo=utc),
        123450,
    )


def test_backup_writes_file_to_output_dir(testconfig, bkp_dir, backup_settings):
    file_content = b"imagefile"
    img1 = os.path.join(testconfig.userimagespath, "1")
    if not os.path.exists(img1):
        os.mkdir(img1)
    with open(os.path.join(img1, "file.txt"), "wb") as f:
        f.write(file_content)

    service = Service(db.session)
    service.create_backup(testconfig, backup_settings)
    assert len(os.listdir(bkp_dir)) == 2  # db and directory
    assert len(os.listdir(os.path.join(bkp_dir, "userimages_backup", "1"))) == 1
    assert os.path.exists(os.path.join(bkp_dir, "userimages_backup", "1", "file.txt"))


def test_timestamp_added_to_db_name(testconfig, bkp_dir, backup_settings):
    assert os.listdir(bkp_dir) == [], "empty dirs at start"
    service = Service(db.session)
    service.create_backup(testconfig, backup_settings)
    dbfile = [f for f in os.listdir(bkp_dir) if f.startswith("lute_backup_2")]
    assert len(dbfile) == 1, "db found"


def test_backup_fails_if_missing_output_dir(testconfig, backup_settings):
    backup_settings.backup_dir = "some_missing_dir"
    service = Service(db.session)
    with pytest.raises(BackupException, match="Missing directory some_missing_dir"):
        service.create_backup(testconfig, backup_settings)


def test_user_can_configure_rolling_backup_count(testconfig, bkp_dir, backup_settings):
    backup_settings.backup_count = 2
    service = Service(db.session)
    for i in range(1, 10):
        service.create_backup(testconfig, backup_settings, suffix=f"0{i}")
    expected_files = [f"lute_backup_0{i}.db.gz" for i in range(8, 10)]
    db_files = [f for f in os.listdir(bkp_dir) if f.endswith(".gz")]
    assert db_files.sort() == expected_files.sort()


def test_all_manual_backups_are_kept(testconfig, bkp_dir, backup_settings):
    backup_settings.backup_count = 2
    service = Service(db.session)
    for i in range(1, 10):
        service.create_backup(
            testconfig, backup_settings, suffix=f"0{i}", is_manual=True
        )
    expected_files = [f"lute_backup_0{i}.db.gz" for i in range(1, 10)]
    db_files = [f for f in os.listdir(bkp_dir) if f.endswith(".gz")]
    assert db_files.sort() == expected_files.sort()


def test_last_import_setting_is_updated_on_successful_backup(
    testconfig, backup_settings
):
    assert backup_settings.last_backup_datetime is None, "no backup"
    service = Service(db.session)
    service.create_backup(testconfig, backup_settings)

    repo = UserSettingRepository(db.session)
    updated = repo.get_backup_settings()
    assert updated.last_backup_datetime is not None, "set"


def test_should_not_run_autobackup_if_auto_is_no_or_false(backup_settings):
    service = Service(db.session)
    backup_settings.backup_enabled = True
    backup_settings.backup_auto = False
    assert service.should_run_auto_backup(backup_settings) is False


def test_autobackup_returns_true_if_never_backed_up(backup_settings):
    backup_settings.backup_enabled = True
    backup_settings.backup_auto = True
    backup_settings.last_backup_datetime = None
    service = Service(db.session)
    assert service.should_run_auto_backup(backup_settings) is True


def test_autobackup_returns_true_if_last_backed_up_over_one_day_ago(backup_settings):
    backup_settings.backup_auto = True
    # backup_settings.backup_warn = True
    curr_datetime = datetime.now()
    one_day_ago = curr_datetime.timestamp() - 24 * 60 * 60

    service = Service(db.session)
    backup_settings.last_backup_datetime = one_day_ago - 10
    assert service.should_run_auto_backup(backup_settings) is True

    backup_settings.last_backup_datetime = one_day_ago + 10
    assert service.should_run_auto_backup(backup_settings) is False


def test_warn_if_last_backup_never_happened_or_is_old(backup_settings):
    curr_datetime = datetime.now()
    one_week_ago = curr_datetime.timestamp() - 7 * 24 * 60 * 60

    backup_settings.backup_warn = True
    backup_settings.last_backup_datetime = None
    service = Service(db.session)

    assert_record_count_equals("select * from books", 0, "sanity check, no books")
    assert service.backup_warning(backup_settings) == "", "no warning if db empty"

    langsvc = LanguageService(db.session)
    langsvc.load_language_def("English")
    assert_record_count_equals("select * from books", 2, "sanity check, have books")
    assert service.backup_warning(backup_settings) == "Never backed up."

    backup_settings.last_backup_datetime = one_week_ago + 10
    assert service.backup_warning(backup_settings) == ""

    backup_settings.last_backup_datetime = one_week_ago - 10
    assert (
        service.backup_warning(backup_settings)
        == "Last backup was more than 1 week ago."
    )

    backup_settings.backup_warn = False
    assert service.backup_warning(backup_settings) == ""


def test_database_backup_file_on_nonexistent_path(bkp_dir):
    file = os.path.join(bkp_dir, "lute_backup_nonexistent.db.gz")
    with pytest.raises(BackupException) as excinfo:
        DatabaseBackupFile(file)
    assert str(excinfo.value) == f"No backup file at {file}."


def test_database_backup_file_with_non_lute_backup(bkp_dir):
    file = os.path.join(bkp_dir, "some_other_file.db.gz")
    with open(file, "wb") as f:
        f.write(b"\0")
    with pytest.raises(BackupException) as excinfo:
        DatabaseBackupFile(file)
    assert str(excinfo.value) == f"Not a valid lute database backup at {file}."


def test_database_backup_file_with_auto_backup_returns_success(auto_backup_file):
    dbf = DatabaseBackupFile(auto_backup_file.path)

    assert auto_backup_file.name == dbf.name
    assert auto_backup_file.path == dbf.filepath
    assert auto_backup_file.size == dbf.size_bytes
    assert dbf.size == "568 KB"
    assert dbf.last_modified == datetime(2024, 1, 1, 0, 0, 0, tzinfo=utc)


def test_database_backup_file_for_auto_is_not_manual(auto_backup_file):
    dbf = DatabaseBackupFile(auto_backup_file.path)
    assert not dbf.is_manual


def test_database_backup_file_for_manual_is_not_manual(manual_backup_file):
    dbf = DatabaseBackupFile(manual_backup_file.path)
    assert dbf.is_manual


def test_database_backup_file_size_formatting(auto_backup_file):
    dbf = DatabaseBackupFile(auto_backup_file.path)

    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 11234567890):
        assert dbf.size == "11 GB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 1123456789):
        assert dbf.size == "1 GB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 19876543):
        assert dbf.size == "20 MB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 7654321):
        assert dbf.size == "8 MB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 1024):
        assert dbf.size == "1 KB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 1001):
        assert dbf.size == "1 KB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 944):
        assert dbf.size == "944 bytes"


def test_list_backups_returns_both(
    bkp_dir, auto_backup_file, manual_backup_file
):  # pylint: disable=unused-argument
    service = Service(db.session)
    backups = service.list_backups(bkp_dir)
    assert len(backups) == 2
    assert all(isinstance(backup, DatabaseBackupFile) for backup in backups)
    assert any(not backup.is_manual for backup in backups)
    assert any(backup.is_manual for backup in backups)


def test_backup_listing_sorts_by_modified(
    bkp_dir, auto_backup_file, manual_backup_file
):  # pylint: disable=unused-argument
    service = Service(db.session)
    backups = service.list_backups(bkp_dir)
    assert len(backups) == 2

    backups.sort()
    assert backups[0].last_modified == datetime(2024, 1, 1, 0, 0, 0, tzinfo=utc)
    assert backups[1].last_modified == datetime(2024, 2, 1, 0, 0, 0, tzinfo=utc)

    backups.sort(reverse=True)
    assert backups[0].last_modified == datetime(2024, 2, 1, 0, 0, 0, tzinfo=utc)
    assert backups[1].last_modified == datetime(2024, 1, 1, 0, 0, 0, tzinfo=utc)



"""
Language service tests.
"""

from lute.language.service import Service
from lute.db import db
from lute.utils.debug_helpers import DebugTimer
from tests.dbasserts import assert_sql_result


def test_get_all_lang_defs(app_context):
    "Can get all predefined languages."
    service = Service(db.session)
    defs = service.get_supported_defs()
    engs = [d for d in defs if d.language.name == "English"]
    assert len(engs) == 1, "have english"
    eng = engs[0]
    assert len(eng.books) == 2, "tutorial and follow-up"
    titles = [b.title for b in eng.books]
    titles.sort()
    assert titles == ["Tutorial", "Tutorial follow-up"], "book titles"


def test_supported_predefined_languages(app_context):
    "Get supported lang names"
    service = Service(db.session)
    predefs = service.supported_predefined_languages()
    assert len(predefs) > 1, "Have predefined"
    langnames = [lang.name for lang in predefs]
    assert "English" in langnames, "Have English"
    assert "French" in langnames, "Have French"


def test_get_language_def(app_context):
    """
    Smoke test, can load a new language from yaml definition.
    """
    service = Service(db.session)
    lang = service.get_language_def("English").language

    assert lang.name == "English"
    assert lang.show_romanization is False, "uses default"
    assert lang.right_to_left is False, "uses default"

    # pylint: disable=line-too-long
    expected = [
        "terms; embeddedhtml; https://simple.wiktionary.org/wiki/[LUTE]; True; 1",
        "terms; popuphtml; https://www.collinsdictionary.com/dictionary/english/[LUTE]; True; 2",
        "sentences; popuphtml; https://www.deepl.com/translator#en/en/[LUTE]; True; 3",
        "terms; popuphtml; https://conjugator.reverso.net/conjugation-english-verb-[LUTE].html; True; 4",
    ]
    actual = [
        f"{ld.usefor}; {ld.dicttype}; {ld.dicturi}; {ld.is_active}; {ld.sort_order}"
        for ld in lang.dictionaries
    ]
    assert actual == expected, "dictionaries"


def test_load_def_loads_lang_and_stories(app_context):
    "Can load a language."
    story_sql = "select bktitle from books order by BkTitle"
    lang_sql = "select LgName from languages"
    assert_sql_result(lang_sql, [], "no langs")
    assert_sql_result(story_sql, [], "nothing loaded")

    dt = DebugTimer("Loading", False)
    dt.step("start")
    service = Service(db.session)
    dt.step("Service()")
    lang_id = service.load_language_def("English")
    dt.step("load_language_def")
    dt.summary()

    assert lang_id > 0, "ID returned, used for filtering"
    assert_sql_result(lang_sql, ["English"], "eng loaded")
    assert_sql_result(story_sql, ["Tutorial", "Tutorial follow-up"], "stories loaded")


def test_load_all_defs_loads_lang_and_stories(app_context):
    "Smoke test, load everything."
    story_sql = "select bktitle from books"
    lang_sql = "select LgName from languages"
    assert_sql_result(lang_sql, [], "no langs")
    assert_sql_result(story_sql, [], "nothing loaded")

    db.session.flush()
    service = Service(db.session)
    defs = service.get_supported_defs()
    langnames = [d.language.name for d in defs]
    for n in langnames:
        lang_id = service.load_language_def(n)
        assert lang_id > 0, "Loaded"



"""
/stats endpoints.
"""

from flask import Blueprint, render_template, jsonify
from lute.stats.service import get_chart_data, get_table_data
from lute.db import db

bp = Blueprint("stats", __name__, url_prefix="/stats")


@bp.route("/")
def index():
    "Main page."
    read_table_data = get_table_data(db.session)
    return render_template("stats/index.html", read_table_data=read_table_data)


@bp.route("/data")
def get_data():
    "Ajax call."
    chartdata = get_chart_data(db.session)
    return jsonify(chartdata)



"""
Flask-wtf forms.
"""

from flask_wtf import FlaskForm
from wtforms import StringField, TextAreaField
from wtforms.validators import DataRequired


class TermTagForm(FlaskForm):
    """
    TermTag.
    """

    text = StringField("Tag", validators=[DataRequired()])
    comment = TextAreaField("Comment")



"""
get_string_indexes tests.
"""

from lute.read.render.calculate_textitems import get_string_indexes


def test_get_string_indexes_scenario(english):
    """
    Run test scenarios.
    """

    zws = "\u200B"
    cases = [
        # Cases are of the following form:
        # ( tokens,
        #   word sought
        #   matches [ [ as_found_in_text, position ], ... ])
        # Finds b
        (["a", "b", "c", "d"], "b", [("b", 1)]),
        # Case doesn't matter
        # The original case is returned
        (["A", "B", "C", "D"], "b", [("b", 1)]),
        # Original case returned
        (["a", "b", "c", "d"], "B", [("b", 1)]),
        # No match
        (["a", "bb", "c", "d"], "B", []),
        # Found in multiple places
        (
            ["b", "b", "c", "d"],
            "b",
            [("b", 0), ("b", 1)],
        ),
        # Multiword, found in multiple
        (
            ["b", "B", "b", "d"],
            f"b{zws}b",
            [(f"b{zws}b", 0), (f"b{zws}b", 1)],
        ),
        # Multiword, found in multiple
        (
            ["b", "B", "c", "b", "b", "x", "b"],
            f"b{zws}b",
            [(f"b{zws}b", 0), (f"b{zws}b", 3)],
        ),
        (["a", " ", "cat", " ", "here"], "cat", [("cat", 2)]),
        (["a", " ", "CAT", " ", "here"], "cat", [("cat", 2)]),
        # No match
        (["a", " ", "CAT", " ", "here"], "ca", []),
        # No match
        (["b", "b", "c", "d"], "x", []),
    ]

    casenum = 0
    for tokens, word, expected in cases:
        casenum += 1
        p = english.parser
        sentence = p.get_lowercase(zws.join(tokens))
        actual = get_string_indexes([p.get_lowercase(word)], sentence)
        msg = f"case {casenum} - tokens: {', '.join(tokens)}; word: {word}"
        assert actual == expected, msg



"""
LangNameParser tests.
"""

# TODO fix names, activate tests.

import pytest

# pylint: disable=wrong-import-order
from lute.models.term import Term
from lute.parse.base import ParsedToken

# TODO fix name
from lute_langname_parser.parser import LangNameParser


def test_dummy_test():
    "A dummy test so that pytest doesn't complain in github ci."
    s = "Hello"
    assert s == "Hello", "TODO - fix these tests for your parser :-)"


# TODO activate tests.
def todo_test_token_count(langname):
    """
    token_count checks.
    """
    cases = [
        ("a", 1),
        ("ab", 1),
        ("ac", 2),
        ("ade", 3),
        ("a_longer_check.", 21),
    ]
    for text, expected_count in cases:
        t = Term(langname, text)
        assert t.token_count == expected_count, text
        assert t.text_lc == t.text, "case"


def assert_tokens_equals(text, lang, expected):
    """
    Parsing a text using a language should give the expected parsed tokens.

    expected is given as array of:
    [ original_text, is_word, is_end_of_sentence ]
    """
    p = LangNameParser()
    actual = p.get_parsed_tokens(text, lang)
    expected = [ParsedToken(*a) for a in expected]
    assert [str(a) for a in actual] == [str(e) for e in expected]


def todo_test_end_of_sentence_stored_in_parsed_tokens(langname):
    """
    ParsedToken is marked as EOS=True at ends of sentences.
    """
    s = "some text。More text? Yep."

    expected = [
        ("你好", True),
        ("。", False, True),
        ("吃饭", True),
        ("了", True),
        ("吗", True),
        ("？", False, True),
        ("现在", True),
        ("是", True),
        ("2024", False, False),
        ("年", True),
        ("。", False, True),
    ]
    assert_tokens_equals(s, langname, expected)


def todo_test_carriage_returns_treated_as_reverse_p_character(langname):
    """
    Returns need to be marked with the backwards P for rendering etc.
    """
    s = "some。\ntext。"

    expected = [
        ("你好", True),
        ("。", False, True),
        ("¶", False, True),
        ("现在", True),
        ("。", False, True),
    ]
    assert_tokens_equals(s, mandarin_chinese, expected)


def todo_test_readings():
    """
    Parser returns readings if they add value.
    """
    p = LangNameParser()

    no_reading = ["Hello"]

    for c in no_reading:
        assert p.get_reading(c) is None, c

    cases = [("你好", "nǐ hǎo"), ("欢迎", "huān yíng"), ("中国", "zhōng guó")]

    for c in cases:
        assert p.get_reading(c[0]) == c[1], c[0]



"""
/bookmarks endpoint
"""

from flask import Blueprint, request, render_template, jsonify
from lute.bookmarks.datatables import get_data_tables_list
from lute.models.book import Text, TextBookmark
from lute.models.repositories import BookRepository
from lute.utils.data_tables import DataTablesFlaskParamParser
from lute.db import db

bp = Blueprint("bookmarks", __name__, url_prefix="/bookmarks")


@bp.route("/<int:bookid>/datatables", methods=["POST"])
def datatables_bookmarks(bookid):
    "Get datatables json for bookmarks."
    parameters = DataTablesFlaskParamParser.parse_params(request.form)
    data = get_data_tables_list(parameters, bookid, db.session)
    return jsonify(data)


@bp.route("/<int:bookid>", methods=["GET"])
def bookmarks(bookid):
    "Get all bookarks for given bookid."
    br = BookRepository(db.session)
    book = br.find(bookid)

    text_dir = "rtl" if book.language.right_to_left else "ltr"
    return render_template("bookmarks/list.html", book=book, text_dir=text_dir)


@bp.route("/add", methods=["POST"])
def add_bookmark():
    "Add bookmark"
    data = request.json
    title = data.get("title")
    try:
        book_id = int(data.get("book_id"))
        page_num = int(data.get("page_num"))
    except ValueError:
        return jsonify(
            success=False, reason="Invalid book_id or page_num provided.", status=200
        )

    if book_id is None or page_num is None or title is None:
        return jsonify(
            success=False,
            reason="Missing value for required parameter 'title' or 'book_id' or page_num.",
            status=200,
        )

    tx = (
        db.session.query(Text)
        .filter(Text.bk_id == book_id)
        .filter(Text.order == page_num)
        .first()
    )
    bookmark = TextBookmark(title=title, text=tx)

    db.session.add(bookmark)
    db.session.commit()
    return jsonify(success=True, status=200)


@bp.route("/delete", methods=["POST"])
def delete_bookmark():
    "Delete bookmark"
    data = request.json
    bookmark_id = None
    try:
        bookmark_id = int(data.get("bookmark_id"))
    except ValueError:
        return jsonify(
            success=False,
            reason=f"Invalid bookmark_id ({data.get('bookmark_id')}) provided.",
            status=200,
        )

    if bookmark_id is None:
        return jsonify(
            success=False,
            reason="Missing required parameter 'bookmark_id'.",
            status=200,
        )

    db.session.query(TextBookmark).filter(TextBookmark.id == bookmark_id).delete()
    db.session.commit()
    return jsonify(success=True, status=200)


@bp.route("/edit", methods=["POST"])
def edit_bookmark():
    "Edit bookmark"
    data = request.json
    bookmark_id = None
    try:
        bookmark_id = int(data.get("bookmark_id"))
    except ValueError:
        return jsonify(
            success=False, reason="Invalid bookmark_id provided.", status=200
        )
    new_title = data.get("new_title", "").strip()

    if bookmark_id is None or new_title == "":
        return jsonify(
            success=False,
            reason="Missing value for required parameter 'new_title' or 'bookmark_id'.",
            status=200,
        )

    db.session.query(TextBookmark).filter(TextBookmark.id == bookmark_id).update(
        {"title": new_title}
    )
    db.session.commit()
    return jsonify(success=True, status=200)



"""
Book tests.
"""

from datetime import datetime
import pytest
from lute.models.language import Language
from lute.book.datatables import get_data_tables_list
from lute.db import db
from lute.db.demo import Service as DemoService
from tests.utils import make_book


@pytest.fixture(name="_dt_params")
def fixture_dt_params():
    "Sample query params."
    columns = [
        {"data": "0", "name": "BkID", "searchable": False, "orderable": False},
        {"data": "1", "name": "BkTitle", "searchable": True, "orderable": True},
        {"data": "2", "name": "IsCompleted", "searchable": False, "orderable": False},
    ]
    params = {
        "draw": "1",
        "columns": columns,
        "order": [{"column": "1", "dir": "asc"}],
        "start": "0",  # Start from page 0
        "length": "10",
        "search": {"value": "", "regex": False},
        "filtLanguage": "0",  # Ha!
    }
    return params


def test_smoke_book_datatables_query_runs(app_context, _dt_params):
    """
    Smoke test only, ensure query runs.
    """
    demosvc = DemoService(db.session)
    demosvc.load_demo_data()
    get_data_tables_list(_dt_params, False, db.session)
    # print(d['data'])
    a = 1
    assert a == 1, "dummy check"


def test_book_query_only_returns_supported_language_books(app_context, _dt_params):
    """
    Smoke test only, ensure query runs.
    """
    demosvc = DemoService(db.session)
    demosvc.load_demo_data()
    for lang in db.session.query(Language).all():
        lang.parser_type = "unknown"
        db.session.add(lang)
    db.session.commit()
    d = get_data_tables_list(_dt_params, False, db.session)
    assert len(d["data"]) == 0, "no books should be active"


def test_book_data_says_completed_if_last_page_has_been_read(
    app_context, _dt_params, english
):
    "Add a visual cue to completed books."
    b = make_book("title", "Hello.", english)
    db.session.add(b)
    db.session.commit()
    _dt_params["search"] = {"value": "title", "regex": False}
    d = get_data_tables_list(_dt_params, False, db.session)
    actual = d["data"][0]
    assert actual["BkID"] == b.id, "correct book"
    assert actual["IsCompleted"] == 0, "not completed"
    t = b.texts[0]
    t.read_date = datetime.now()
    db.session.add(t)
    db.session.commit()
    d = get_data_tables_list(_dt_params, False, db.session)
    actual = d["data"][0]
    assert actual["BkID"] == b.id, "correct book"
    assert actual["IsCompleted"] == 1, "completed"



"Debug helpers."

import time


class DebugTimer:
    """
    Helper to log time.
    """

    global_step_map = {}

    def __init__(self, name, display=True):
        self.start = time.process_time()
        self.curr_start = self.start
        self.name = name
        self.step_map = {}
        self.display = display
        if display:
            print(f"{name} timer started")

    def step(self, s):
        "Dump time spent in step, total time since start."
        n = time.process_time()
        step_elapsed = n - self.curr_start
        total_step_elapsed = self.step_map.get(s, 0)
        total_step_elapsed += step_elapsed
        self.step_map[s] = total_step_elapsed

        if s != "":
            full_step_map_string = f"{self.name} {s}"
            global_step_elapsed = DebugTimer.global_step_map.get(
                full_step_map_string, 0
            )
            global_step_elapsed += step_elapsed
            DebugTimer.global_step_map[full_step_map_string] = global_step_elapsed

        total_elapsed = n - self.start
        self.curr_start = n

        if not self.display:
            return

        msg = " ".join(
            [
                f"{self.name} {s}:",
                f"step_elapsed: {step_elapsed:.6f},",
                f"total step_elapsed: {total_step_elapsed:.6f},",
                f"total_elapsed: {total_elapsed:.6f}",
            ]
        )
        print(msg, flush=True)

    def summary(self):
        "Print final step summary."
        print(f"{self.name} summary ------------------", flush=True)
        for k, v in self.step_map.items():
            print(f"  {k}: {v:.6f}", flush=True)
        print(f"end {self.name} summary --------------", flush=True)

    @classmethod
    def clear_total_summary(cls):
        cls.global_step_map = {}

    @classmethod
    def total_summary(cls):
        "Print final step summary."
        print("global summary ------------------", flush=True)
        for k, v in cls.global_step_map.items():
            print(f"  {k}: {v:.6f}", flush=True)
        print("end global summary --------------", flush=True)



"Language helper methods."

import os
import re
from glob import glob
import yaml
from lute.models.language import Language
from lute.book.model import Book, Repository

# from lute.utils.debug_helpers import DebugTimer


class LangDef:
    "Language, built from language definition.yml, and .txt book files."

    # Map of definition.yaml directory to the yaml.safe_load content.
    # The definition files never change, and loading them takes time, so
    # cache it for better unit test performance.
    yaml_cache = {}

    @classmethod
    def _get_loaded_yaml(cls, definition_file_path):
        "Get from cache, or load it and return it."
        if definition_file_path not in LangDef.yaml_cache:
            with open(definition_file_path, "r", encoding="utf-8") as df:
                d = yaml.safe_load(df)
                LangDef.yaml_cache[definition_file_path] = d
        return LangDef.yaml_cache[definition_file_path]

    def __init__(self, directory):
        "Build from files."
        self.directory = directory
        self.language_name = self._get_name(directory)

    def _get_name(self, directory):
        def_file = os.path.join(directory, "definition.yaml")
        d = LangDef._get_loaded_yaml(def_file)
        return d["name"]

    @property
    def language(self):
        return self._load_lang_def(self.directory)

    @property
    def books(self):
        return self._get_books(self.directory, self.language_name)

    def _load_lang_def(self, directory):
        "Load from file, must exist."
        def_file = os.path.join(directory, "definition.yaml")
        d = LangDef._get_loaded_yaml(def_file)
        return Language.from_dict(d)

    def _get_books(self, directory, language_name):
        "Get the stories."
        books = []
        story_glob = os.path.join(directory, "*.txt")
        for filename in glob(story_glob):
            with open(filename, "r", encoding="utf-8") as f:
                content = f.read()
            title_match = re.search(r"title:\s*(.*)\n", content)
            title = title_match.group(1).strip()
            content = re.sub(r"#.*\n", "", content)
            b = Book()
            b.language_name = language_name
            b.title = title
            b.text = content
            books.append(b)
        return books


class Service:
    "Service."

    def __init__(self, session):
        self.session = session
        self.lang_defs_cache = self._get_langdefs_cache()

    def _get_langdefs_cache(self):
        "Load cache."
        # dt = DebugTimer("_get_langdefs_cache", False)
        # dt.step("start")
        thisdir = os.path.dirname(__file__)
        langdefs_dir = os.path.join(thisdir, "..", "db", "language_defs")
        langdefs_dir = os.path.abspath(langdefs_dir)
        # dt.step("got base directory")
        cache = []
        def_glob = os.path.join(langdefs_dir, "**", "definition.yaml")
        def_list = glob(def_glob)
        # dt.step("globbed")
        def_list.sort()
        for f in def_list:
            lang_dir, _ = os.path.split(f)
            ld = LangDef(lang_dir)
            # dt.step(f"build ld {ld.language_name}".ljust(30))
            cache.append(ld)
        # dt.summary()
        return cache

    def get_supported_defs(self):
        "Return supported language definitions."
        ret = [ld for ld in self.lang_defs_cache if ld.language.is_supported]
        ret.sort(key=lambda x: x.language_name)
        return ret

    def supported_predefined_languages(self):
        "Supported Languages defined in yaml files."
        return [d.language for d in self.get_supported_defs()]

    def get_language_def(self, lang_name):
        "Get a lang def and its stories."
        ret = [ld for ld in self.lang_defs_cache if ld.language_name == lang_name]
        if len(ret) == 0:
            raise RuntimeError(f"Missing language def name {lang_name}")
        return ret[0]

    def load_language_def(self, lang_name):
        "Load a language def and its stories, save to database."
        load_def = self.get_language_def(lang_name)
        lang = load_def.language
        if not lang.is_supported:
            raise RuntimeError(f"{lang_name} not supported, can't be loaded.")

        self.session.add(lang)
        self.session.commit()

        r = Repository(self.session)
        for b in load_def.books:
            r.add(b)
        r.commit()

        return lang.id



"""
Current settings tests.
"""

from lute.db import db
from lute.settings.current import refresh_global_settings, current_settings


def test_refresh_refreshes_current_settings(app_context):
    "Current settigns are loaded."
    if "backup_dir" in current_settings:
        del current_settings["backup_dir"]
    refresh_global_settings(db.session)
    assert "backup_dir" in current_settings, "loaded"



"""
Db management.
"""

import os
from sqlalchemy import text
from flask import current_app
from lute.models.setting import UserSetting
from lute.models.repositories import UserSettingRepository


def delete_all_data(session):
    """
    DANGEROUS!  Delete everything, restore user settings, clear sys settings.

    NO CHECKS ARE PERFORMED.
    """

    # Setting the pragma first ensures cascade delete.
    statements = [
        "pragma foreign_keys = ON",
        "delete from languages",
        "delete from tags",
        "delete from tags2",
        "delete from settings",
    ]
    for s in statements:
        session.execute(text(s))
    session.commit()
    add_default_user_settings(session, current_app.env_config.default_user_backup_path)


def _revised_mecab_path(repo):
    """
    Change the mecab_path if it's not found, and a
    replacement is found.

    Lute Docker images are built to be multi-arch, and
    interestingly (annoyingly), mecab libraries are installed into
    different locations depending on the architecture, even with
    the same Dockerfile and base image.

    Returns: new mecab path if old one is missing _and_
    new one found, otherwise just return the old one.
    """

    mp = repo.get_value("mecab_path")
    if mp is not None and os.path.exists(mp):
        return mp

    # See develop docs for notes on how to find the libmecab path!
    candidates = [
        # linux/arm64
        "/lib/aarch64-linux-gnu/libmecab.so.2",
        # linux/amd64
        "/lib/x86_64-linux-gnu/libmecab.so.2",
        # github CI, ubuntu-latest
        "/lib/x86_64-linux-gnu/libmecab.so.2",
    ]
    replacements = [p for p in candidates if os.path.exists(p)]
    if len(replacements) > 0:
        return replacements[0]
    # Replacement not found, leave current value as-is.
    return mp


def add_default_user_settings(session, default_user_backup_path):
    """
    Load missing user settings with default values.
    """
    repo = UserSettingRepository(session)

    # These keys are rendered into the global javascript namespace var
    # LUTE_USER_SETTINGS, so if any of these keys change, check the usage
    # of that variable as well.
    keys_and_defaults = {
        "backup_enabled": True,
        "backup_auto": True,
        "backup_warn": True,
        "backup_dir": default_user_backup_path,
        "backup_count": 5,
        "lastbackup": None,
        "mecab_path": None,
        "japanese_reading": "hiragana",
        "current_theme": "-",
        "custom_styles": "/* Custom css to modify Lute's appearance. */",
        "show_highlights": True,
        "current_language_id": 0,
        # Behaviour:
        "open_popup_in_new_tab": False,
        "stop_audio_on_term_form_open": True,
        "stats_calc_sample_size": 5,
        # Keyboard shortcuts.  These have default values assigned
        # as they were the hotkeys defined in the initial Lute
        # release.
        "hotkey_StartHover": "escape",
        "hotkey_PrevWord": "arrowleft",
        "hotkey_NextWord": "arrowright",
        "hotkey_StatusUp": "arrowup",
        "hotkey_StatusDown": "arrowdown",
        "hotkey_Bookmark": "b",
        "hotkey_CopySentence": "c",
        "hotkey_CopyPara": "shift+c",
        "hotkey_TranslateSentence": "t",
        "hotkey_TranslatePara": "shift+t",
        "hotkey_NextTheme": "m",
        "hotkey_ToggleHighlight": "h",
        "hotkey_ToggleFocus": "f",
        "hotkey_Status1": "1",
        "hotkey_Status2": "2",
        "hotkey_Status3": "3",
        "hotkey_Status4": "4",
        "hotkey_Status5": "5",
        "hotkey_StatusIgnore": "i",
        "hotkey_StatusWellKnown": "w",
        "hotkey_SaveTerm": "ctrl+enter",
        # New hotkeys.  These must have empty values, because
        # users may have already setup their hotkeys, and we can't
        # assume that a given key combination is free:
        "hotkey_CopyPage": "",
        "hotkey_DeleteTerm": "",
        "hotkey_EditPage": "",
        "hotkey_TranslatePage": "",
        "hotkey_PrevUnknownWord": "",
        "hotkey_NextUnknownWord": "",
        "hotkey_PrevSentence": "",
        "hotkey_NextSentence": "",
    }
    for k, v in keys_and_defaults.items():
        if not repo.key_exists(k):
            s = UserSetting()
            s.key = k
            s.value = v
            session.add(s)
    session.commit()

    # Revise the mecab path if necessary.
    # Note this is done _after_ the defaults are loaded,
    # because the user may have already loaded the defaults
    # (e.g. on machine upgrade) and stored them in the db,
    # so we may have to _update_ the existing setting.
    revised_mecab_path = _revised_mecab_path(repo)
    repo.set_value("mecab_path", revised_mecab_path)
    session.commit()



"""
Term import.
"""

import csv

from lute.models.term import Status
from lute.models.repositories import LanguageRepository
from lute.term.model import Term, Repository


class BadImportFileError(Exception):
    """
    Raised if the import file is bad:

    - unknown language or status
    - dup term
    etc.
    """


class Service:
    "Service."

    def __init__(self, session):
        self.session = session

    def import_file(
        self, filename, create_terms=True, update_terms=True, new_as_unknowns=False
    ):
        """
        Validate and import file.

        Throws BadImportFileError if file contains invalid data.
        """
        import_data = self._load_import_file(filename)
        self._validate_data(import_data)
        return self._do_import(import_data, create_terms, update_terms, new_as_unknowns)

    def _load_import_file(self, filename, encoding="utf-8-sig"):
        "Create array of hashes from file."
        unique_rows = set()
        importdata = []
        with open(filename, "r", encoding=encoding) as f:
            reader = csv.DictReader(f)
            if reader.fieldnames:  # Avoid empty file error
                reader.fieldnames = [name.lower() for name in reader.fieldnames]

            fieldnames = reader.fieldnames
            if fieldnames is None:
                raise BadImportFileError("No terms in file")
            self._validate_data_fields(fieldnames)

            line_num = 0
            for line in reader:
                line_num += 1
                if None in line.values():
                    raise BadImportFileError(f"Missing values on line {line_num}")
                if None in line.keys():
                    raise BadImportFileError(f"Extra values on line {line_num}")
                line_tuple = tuple(line.items())
                if line_tuple not in unique_rows:
                    unique_rows.add(line_tuple)
                    importdata.append(line)

        if len(importdata) == 0:
            raise BadImportFileError("No terms in file")

        return importdata

    def _validate_data_fields(self, field_list):
        "Check the keys in the file."
        required = ["language", "term"]
        for k in required:
            if k not in field_list:
                raise BadImportFileError(f"Missing required field '{k}'")

        allowed = required + [
            "translation",
            "parent",
            "status",
            "tags",
            "pronunciation",
            "link_status",
        ]
        ignored = ["added"]
        for k in field_list:
            if k not in allowed and k not in ignored:
                raise BadImportFileError(f"Unknown field '{k}'")

    def _validate_data(self, import_data):
        """
        Check the data.
        """
        self._validate_languages(import_data)
        self._validate_terms_exist(import_data)
        self._validate_statuses(import_data)
        self._validate_no_duplicate_terms(import_data)

    def _create_langs_dict(self, import_data):
        "Create dictionary of language name to Language."
        repo = LanguageRepository(self.session)
        lang_dict = {}
        langs = [hsh["language"].strip() for hsh in import_data]
        for lang_name in list(set(langs)):
            lang_dict[lang_name] = repo.find_by_name(lang_name)
        return lang_dict

    def _get_status(self, s):
        "Convert status to db value."
        status_map = {
            "": 1,
            "1": 1,
            "2": 2,
            "3": 3,
            "4": 4,
            "5": 5,
            "W": Status.WELLKNOWN,
            "I": Status.IGNORED,
        }
        return status_map.get(s)

    def _validate_languages(self, import_data):
        "Validate the languages."
        lang_dict = self._create_langs_dict(import_data)
        for lang_name, lang in lang_dict.items():
            if lang is None:
                raise BadImportFileError(f"Unknown language '{lang_name}'")

    def _validate_statuses(self, import_data):
        "All statuses must be valid."
        statuses = [hsh["status"].strip() for hsh in import_data if "status" in hsh]
        for s in set(statuses):
            if self._get_status(s) is None:
                raise BadImportFileError(
                    "Status must be one of 1, 2, 3, 4, 5, I, W, or blank"
                )

    def _validate_terms_exist(self, import_data):
        "All records must have a term."
        blanks = [hsh for hsh in import_data if hsh["term"].strip() == ""]
        if len(blanks) > 0:
            raise BadImportFileError("Term is required")

    def _validate_no_duplicate_terms(self, import_data):
        """
        Duplicate terms aren't allowed.

        If file contained two duplicate terms, which is the "correct" one?
        """

        def make_lang_term_string(hsh):
            t = hsh["term"].strip()
            # Have to also clear unicode whitespace.
            t = " ".join(t.split())
            return f"{hsh['language']}: {t.lower()}"

        lang_terms = [make_lang_term_string(hsh) for hsh in import_data]
        term_counts = {}
        for term in lang_terms:
            term_counts[term] = term_counts.get(term, 0) + 1
        duplicates = [term for term, count in term_counts.items() if count > 1]
        if len(duplicates) != 0:
            raise BadImportFileError(
                f"Duplicate terms in import: {', '.join(duplicates)}"
            )

    def _import_term_skip_parents(self, repo, rec, lang, set_to_unknown=False):
        "Add a single record to the repo."
        t = Term()
        t.language = lang
        t.language_id = lang.id
        t.text = rec["term"]
        if "translation" in rec:
            t.translation = rec["translation"]
        if "status" in rec:
            status = self._get_status(rec["status"])
            if status is not None:
                t.status = int(status)
        if set_to_unknown:
            t.status = 0
        if "pronunciation" in rec:
            t.romanization = rec["pronunciation"]
        if "tags" in rec:
            tags = list(map(str.strip, rec["tags"].split(",")))
            t.term_tags = [t for t in tags if t != ""]
        repo.add(t)

    def _update_term_skip_parents(self, t, repo, rec):
        "Update a term in the repo."
        # Don't change the lang or text of the term
        # t.language = lang
        # t.language_id = lang.id
        # t.text = rec["term"]
        if "translation" in rec:
            t.translation = rec["translation"]
        if "status" in rec:
            status = self._get_status(rec["status"])
            if status is not None:
                t.status = int(status)
        if "pronunciation" in rec:
            t.romanization = rec["pronunciation"]
        if "tags" in rec:
            tags = list(map(str.strip, rec["tags"].split(",")))
            t.term_tags = [t for t in tags if t != ""]

        repo.add(t)

    def _set_term_parents(self, repo, rec, lang):
        "Set the term parents."
        t = repo.find(lang.id, rec["term"])
        parents = list(map(str.strip, rec["parent"].split(",")))
        t.parents = [p for p in parents if p != ""]
        if "link_status" in rec and len(parents) == 1:
            sync_status = rec["link_status"] or ""
            t.sync_status = sync_status.strip().lower() == "y"
        if len(t.parents) != 1:
            t.sync_status = False

        # If syncing to parent, and the term status was not explicitly set,
        # then "inherit" the parent status.
        if t.sync_status and len(t.parents) == 1 and "status" not in rec:
            p = repo.find(lang.id, t.parents[0])
            if p is not None:
                t.status = p.status

        repo.add(t)

    def _do_import(
        self, import_data, create_terms=True, update_terms=True, new_as_unknowns=False
    ):
        """
        Import records.

        If create_terms is True, create new terms.
        If update_terms is True, update existing terms.
        If new_as_unknowns is True, new terms are given status 0.

        The import is done in two passes:
        1. import the basic terms, without setting their parents
        2. update the terms with parents

        The two passes are done because the import file may
        contain a parent in its own row, and we want that to be
        imported first to get its own specified data.
        """
        # pylint: disable=too-many-locals

        repo = Repository(self.session)

        skipped = 0

        # Keep track of the created and updated terms: we only want to
        # update these ones in pass #2.
        created_terms = []
        updated_terms = []

        def term_string(lang, term):
            return f"{lang.id}-{term}"

        for batch in [
            import_data[i : i + 100] for i in range(0, len(import_data), 100)
        ]:
            langs_dict = self._create_langs_dict(batch)
            for hsh in batch:
                lang = langs_dict[hsh["language"]]
                t = repo.find(lang.id, hsh["term"])
                ts = term_string(lang, hsh["term"])

                if create_terms and t is None:
                    # Create a brand-new term.
                    self._import_term_skip_parents(repo, hsh, lang, new_as_unknowns)
                    created_terms.append(ts)

                elif update_terms and t is not None:
                    # Can only update existing terms.
                    self._update_term_skip_parents(t, repo, hsh)
                    updated_terms.append(ts)

                else:
                    skipped += 1

            repo.commit()

        pass_2 = [t for t in import_data if "parent" in t and t["parent"] != ""]
        for batch in [pass_2[i : i + 100] for i in range(0, len(pass_2), 100)]:
            langs_dict = self._create_langs_dict(batch)
            for hsh in batch:
                lang = langs_dict[hsh["language"]]
                ts = term_string(lang, hsh["term"])
                if ts in created_terms or ts in updated_terms:
                    self._set_term_parents(repo, hsh, lang)
            repo.commit()

        stats = {
            "created": len(created_terms),
            "updated": len(updated_terms),
            "skipped": skipped,
        }

        return stats



"""
User entry point.

Start lute running on given port, or 5001 if not set.

e.g.

python -m lute.main --port 5001
"""
import errno
import os
import argparse
import shutil
import logging
import textwrap
from waitress import serve
from lute import __version__
from lute.app_factory import create_app
from lute.config.app_config import AppConfig
from lute.db import db
from lute.db.demo import Service as DemoService

logging.getLogger("waitress.queue").setLevel(logging.ERROR)
logging.getLogger("natto").setLevel(logging.CRITICAL)


def _print(s):
    """
    Print message to stdout.
    """
    if isinstance(s, str):
        s = s.split("\n")
    msg = "\n".join(f"  {lin}" for lin in s)
    print(msg, flush=True)


def _create_prod_config_if_needed():
    """
    If config.yml is missing, create one from prod config.
    """
    config_file = os.path.join(AppConfig.configdir(), "config.yml")
    if not os.path.exists(config_file):
        prod_conf = os.path.join(AppConfig.configdir(), "config.yml.prod")
        shutil.copy(prod_conf, config_file)
        _print(["", "Using new production config.", ""])


def _get_config_file_path(config_file_path=None):
    """
    Get final config file to use.

    Uses config file if set (throws if doesn't exist);
    otherwise, uses the prod config, creating a prod config
    if necessary.
    """
    use_config = config_file_path
    if config_file_path is not None:
        _print(f"Using specified config: {config_file_path}")
    elif os.path.exists("config.yml"):
        _print("Using config.yml found in root")
        use_config = "config.yml"
    else:
        _print("Using default config")
        _create_prod_config_if_needed()
        use_config = AppConfig.default_config_filename()

    ac = AppConfig(use_config)
    _print(f"  data path: {ac.datapath}")
    _print(f"  database:  {ac.dbfilename}")
    if ac.is_docker:
        _print("  (Note these are container paths, not host paths.)")
    _print("")

    return use_config


def _start(args):
    "Configure and start the app."
    _print(f"\nStarting Lute version {__version__}.\n")

    config_file_path = _get_config_file_path(args.config)
    app = create_app(config_file_path, output_func=_print)
    with app.app_context():
        demosvc = DemoService(db.session)
        if demosvc.should_load_demo_data():
            _print("Loading demo data.")
            demosvc.load_demo_data()

    close_msg = """
    When you're finished reading, stop this process
    with Ctrl-C or your system equivalent.
    """
    if app.env_config.is_docker:
        close_msg = """
        When you're finished reading, stop this container
        with Ctrl-C, docker compose stop, or docker stop <containerid>
        as appropriate.
        """
    _print(textwrap.dedent(close_msg))

    host_ip = "127.0.0.1" if args.local else "0.0.0.0"
    ip_port = f"{host_ip}:{args.port}"
    msg = f"""Lute v{__version__} is running on {ip_port}.  Open a web browser and go to:

    http://localhost:{args.port}
    """
    _print(textwrap.dedent(msg))

    try:
        serve(app, host=host_ip, port=args.port)
    except OSError as err:
        if err.errno == errno.EADDRINUSE:
            msg = [
                f"ERROR: port {args.port} is already in use.",
                "please try adding a --port parameter, e.g.:",
                "",
                "  python -m lute.main --port 9876",
                "",
            ]
            _print(msg)
        else:
            # Throw back up, to get general error message
            raise


def start():
    "Main entry point.  Called via scripts and pyproject.toml."
    parser = argparse.ArgumentParser(description="Start lute.")
    parser.add_argument(
        "--local",
        action="store_true",
        help="Run local only (not accessible on other devices on the same network)",
    )
    parser.add_argument(
        "--port", type=int, default=5001, help="Port number (default: 5001)"
    )
    parser.add_argument(
        "--config",
        help="Path to override config file.  Uses lute/config/config.yml if not set.",
    )
    try:
        _start(parser.parse_args())
    except Exception as e:  # pylint: disable=broad-exception-caught
        dashes = "-" * 50
        failmsg = f"""
        {dashes}
        Error during startup:
        Type: {type(e)}
        {e}

        Please check your setup and try again.
        Ask for help on Discord, or report an issue on GitHub.
        Additionally, help is available with --help.
        {dashes}
        """

        print(textwrap.dedent(failmsg))


if __name__ == "__main__":
    start()



"""
Acceptance test fixtures and step definitions.

Fixtures used in most or all tests:
_environment_check: ensures site is up
chromebrowser: creates a properly configured browser
luteclient: WIPES THE DB and provides helpful wrappers
"""

import os
import re
import tempfile
import time
import yaml
import requests

import pytest
from pytest_bdd import given, when, then, parsers
from selenium.webdriver.chrome.options import Options as ChromeOptions
from splinter import Browser
from tests.acceptance.lute_test_client import LuteTestClient


def pytest_addoption(parser):
    """
    Command-line args for pytest runs.
    """
    parser.addoption("--port", action="store", type=int, help="Specify the port number")
    parser.addoption("--headless", action="store_true", help="Run the test as headless")


@pytest.fixture(name="_environment_check", scope="session")
def fixture_env_check(request):
    """
    Sanity check that the site is up, port is set.
    """
    useport = request.config.getoption("--port")
    if useport is None:
        # Need to specify the port, e.g.
        # pytest tests/acceptance --port=1234
        # Acceptance tests run using 'inv accept' sort this out automatically.
        pytest.exit("--port not set")

    # Try connecting a few times.
    success = False
    max_attempts = 5
    curr_attempt = 0
    url = f"http://localhost:{useport}/"

    while curr_attempt < max_attempts and not success:
        curr_attempt += 1
        try:
            requests.get(url, timeout=10)
            success = True
        except requests.exceptions.ConnectionError:
            time.sleep(5)

    if not success:
        msg = f"Unable to reach {url} after {curr_attempt} tries ... "
        msg += "is it running?  Use inv accept to auto-start it."
        pytest.exit(msg)
        print()
    else:
        print(f"Connected successfully after {curr_attempt} tries")


@pytest.fixture(name="chromebrowser", scope="session")
def session_chrome_browser(request, _environment_check):
    """
    Create a chrome browser.

    For some weird reason, this performs **MUCH**
    better than the default "browser" fixture provided by
    splinter/pytest-splinter:

    "with self.browser.get_iframe('wordframe') as iframe"
      - Without this custom fixture: 5+ seconds!
      - With this fixture: 0.03 seconds

    The times were consistent with various options: headless,
    non, virus scanning on/off, etc.
    """
    chrome_options = ChromeOptions()

    headless = request.config.getoption("--headless")
    if headless:
        chrome_options.add_argument("--headless")  # Enable headless mode

        # Set the window size and ensure no devtools, or errors happen:
        #
        # selenium.common.exceptions.ElementClickInterceptedException:
        # Message: element click intercepted: Element <button id="submit" ... >
        #   is not clickable at poi...
        #
        # Possibly running headless is opening devtools or using
        # a smaller browser window, which affects the layout and
        # hides some elements.  When run in non-headless, all is fine.

        # https://stackoverflow.com/questions/54023497/
        #   python-selenium-chrome-driver-disable-devtools
        chrome_options.add_experimental_option("excludeSwitches", ["enable-logging"])

        # https://stackoverflow.com/questions/43541925/
        #   how-can-i-set-the-browser-window-size-when-using-google-chrome-headless
        chrome_options.add_argument("window-size=1920,1080")

    # Initialize the browser with ChromeOptions
    browser = Browser("chrome", options=chrome_options)

    # Set up and clean up the browser
    def fin():
        browser.quit()

    request.addfinalizer(fin)

    return browser


@pytest.fixture(name="luteclient")
def fixture_lute_client(request, chromebrowser):
    """
    Start the lute browser.
    """
    useport = request.config.getoption("--port")
    url = f"http://localhost:{useport}"
    c = LuteTestClient(chromebrowser, url)
    yield c


@pytest.fixture(name="_restore_jp_parser")
def fixture_restore_jp_parser(luteclient):
    "Hack for test: restore a parser using the dev api."
    yield
    luteclient.change_parser_registry_key("disabled_japanese", "japanese")


################################
## STEP DEFS


# Setup


@when(parsers.parse("sleep for {seconds}"))
def _sleep(seconds):
    "Hack helper."
    time.sleep(int(seconds))


@given("a running site")
def given_running_site(luteclient):
    "Sanity check!"
    resp = requests.get(luteclient.home, timeout=5)
    assert resp.status_code == 200, f"{luteclient.home} is up"
    luteclient.visit("/")
    luteclient.clear_book_filter()
    assert luteclient.browser.is_text_present("Lute")


@given('I disable the "japanese" parser')
def disable_japanese_parser(luteclient, _restore_jp_parser):
    luteclient.change_parser_registry_key("japanese", "disabled_japanese")


@given('I enable the "japanese" parser')
def enable_jp_parser(luteclient):
    luteclient.change_parser_registry_key("disabled_japanese", "japanese")


# Browsing


@given(parsers.parse('I visit "{p}"'))
def given_visit(luteclient, p):
    "Go to a page."
    luteclient.visit(p)


@when(parsers.parse('I click the "{linktext}" link'))
def when_click_link_text(luteclient, linktext):
    luteclient.browser.links.find_by_text(linktext).click()


@then(parsers.parse("the page title is {title}"))
def then_title(luteclient, title):
    assert luteclient.browser.title == title


@then(parsers.parse('the page contains "{text}"'))
def then_page_contains(luteclient, text):
    assert text in luteclient.browser.html


# Language


@given("demo languages")
def given_demo_langs_loaded(luteclient):
    "Spot check some things exist."
    for s in ["English", "Spanish"]:
        assert s in luteclient.language_ids, f"Check map for {s}"


@given("the demo stories are loaded")
def given_demo_stories_loaded(luteclient):
    "Load the demo stories."
    luteclient.load_demo_stories()
    luteclient.visit("/")
    luteclient.clear_book_filter()


@given("I clear the book filter")
def given_clear_book_filter(luteclient):
    "clear filter."
    luteclient.visit("/")
    luteclient.clear_book_filter()


@given(parsers.parse("I update the {lang} language:\n{content}"))
def given_update_language(luteclient, lang, content):
    "Content is assumed to be yaml."
    updates = yaml.safe_load(content)
    luteclient.edit_language(lang, updates)


# Books


@given(parsers.parse('a {lang} book "{title}" with content:\n{c}'))
def given_book(luteclient, lang, title, c):
    "Make a book."
    luteclient.make_book(title, c, lang)
    _sleep(1)  # Hack!


@given(parsers.parse('a {lang} book "{title}" from file {filename}'))
def given_book_from_file(luteclient, lang, title, filename):
    "Book is made from file in sample_files dir."
    thisdir = os.path.dirname(os.path.realpath(__file__))
    fullpath = os.path.join(thisdir, "sample_files", filename)
    luteclient.make_book_from_file(title, fullpath, lang)
    _sleep(1)  # Hack!


@given(parsers.parse("a {lang} book from url {url}"))
def given_book_from_url(luteclient, lang, url):
    "Book is made from url in dev_api."
    luteclient.make_book_from_url(url, lang)
    _sleep(1)  # Hack!


@given(parsers.parse('the book table loads "{title}"'))
def given_book_table_wait(luteclient, title):
    "The book table is loaded via ajax, so there's a delay."
    _sleep(1)  # Hack!
    assert title in luteclient.browser.html


@when(parsers.parse('I set the book table filter to "{filt}"'))
def when_set_book_table_filter(luteclient, filt):
    "Set the filter, wait a sec."
    b = luteclient.browser
    b.find_by_tag("input").fill(filt)
    time.sleep(1)


@then(parsers.parse("the book table contains:\n{content}"))
def check_book_table(luteclient, content):
    "Check the table, e.g. content like 'Hola; Spanish; ; 4 (0%);'"
    time.sleep(1)
    assert content == luteclient.get_book_table_content()


# Terms


@given(parsers.parse("a new {lang} term:\n{content}"))
def given_new_term(luteclient, lang, content):
    "The content is assumed to be yaml."
    updates = yaml.safe_load(content)
    luteclient.make_term(lang, updates)


@given(parsers.parse("import term file:\n{content}"))
def import_term_file(luteclient, content):
    "Import the term file."
    luteclient.visit("/")
    luteclient.browser.find_by_css("#menu_terms").mouse_over()
    luteclient.browser.find_by_id("term_import_index").first.click()
    fd, path = tempfile.mkstemp()
    with os.fdopen(fd, "w") as tmp:
        # do stuff with temp file
        tmp.write(content)
    luteclient.browser.attach_file("text_file", path)
    luteclient.browser.find_by_id("create_terms").click()
    luteclient.browser.find_by_id("update_terms").click()
    luteclient.browser.find_by_id("btnSubmit").click()


@then(parsers.parse("the term table contains:\n{content}"))
def check_term_table(luteclient, content):
    "Check the table."
    luteclient.visit("/")
    luteclient.browser.find_by_css("#menu_terms").mouse_over()
    luteclient.browser.find_by_id("term_index").first.click()
    time.sleep(1)
    if content == "-":
        content = "No data available in table"
    assert content == luteclient.get_term_table_content()


@when("click Export CSV")
def click_export_csv(luteclient):
    "Export the term csv"
    luteclient.browser.find_by_css("#term_actions").mouse_over()
    luteclient.click_link("Export CSV")


@then(parsers.parse("exported CSV file contains:\n{content}"))
def check_exported_file(luteclient, content):
    "Check the exported file, replace all dates with placeholder."
    actual = luteclient.get_temp_file_content("export_terms.csv").strip()
    actual = re.sub(r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}", "DATE_HERE", actual)
    assert content == actual


# Reading


@then(parsers.parse("the reading pane shows:\n{content}"))
def then_read_content(luteclient, content):
    "Check rendered content."
    displayed = luteclient.displayed_text()
    assert content.replace("\n", "/") == displayed


@when(parsers.parse("I change the current text content to:\n{content}"))
def when_change_content(luteclient, content):
    "Change the content."
    assert "Reading" in luteclient.browser.title, "sanity check"
    b = luteclient.browser
    b.find_by_css("div.hamburger-btn").first.click()
    b.find_by_id("page-operations-title").click()
    b.find_by_id("editText").click()
    b.find_by_id("text").fill(content)
    b.find_by_id("submit").click()


@when(parsers.parse("I add a page {position} current with content:\n{content}"))
def when_add_page(luteclient, position, content):
    "Change the content."
    assert "Reading" in luteclient.browser.title, "sanity check"
    b = luteclient.browser
    b.find_by_css("div.hamburger-btn").first.click()
    b.find_by_id("page-operations-title").click()

    assert position in ["before", "after"], "sanity check"
    linkid = "readmenu_add_page_before"
    if position == "after":
        linkid = "readmenu_add_page_after"
    b.find_by_id(linkid).click()
    b.find_by_id("text").fill(content)
    b.find_by_id("submit").click()
    b.reload()


@when(parsers.parse("I go to the {position} page"))
def when_go_to_page(luteclient, position):
    "Go to page."
    assert "Reading" in luteclient.browser.title, "sanity check"
    assert position in ["previous", "next"], "sanity check"

    linkid = "navNext"
    if position == "previous":
        linkid = "navPrev"
    b = luteclient.browser
    b.find_by_id(linkid).first.click()
    time.sleep(0.1)  # Assume this is necessary for ajax reload.
    # Don't reload, as it seems to nullify the nav click.
    # b.reload()


@when(parsers.parse("I delete the current page"))
def when_delete_current_page(luteclient):
    "Delete the current page."
    assert "Reading" in luteclient.browser.title, "sanity check"
    b = luteclient.browser
    b.find_by_css("div.hamburger-btn").first.click()
    b.find_by_id("page-operations-title").click()
    b.find_by_id("readmenu_delete_page").first.click()
    alert = b.get_alert()
    alert.accept()
    b.reload()


# Reading, terms


@when(parsers.parse('I click "{word}" and edit the form:\n{content}'))
def when_click_word_edit_form(luteclient, word, content):
    "The content is assumed to be yaml."
    updates = yaml.safe_load(content)
    luteclient.click_word_fill_form(word, updates)


@then(parsers.parse('the reading page term form frame contains "{text}"'))
def then_reading_page_term_form_iframe_contains(luteclient, text):
    "Have to get and read the iframe content, it's not in the main browser page."
    with luteclient.browser.get_iframe("wordframe") as iframe:
        assert text in iframe.html


# Reading, word actions


@when(parsers.parse('I click "{word}"'))
def when_click_word(luteclient, word):
    "Click word."
    luteclient.click_word(word)


@when(parsers.parse("I shift click:\n{words}"))
def shift_click_terms(luteclient, words):
    "Shift-click"
    words = words.split("\n")
    luteclient.shift_click_words(words)


@when(parsers.parse('I click "{word}" and press hotkey "{hotkey}"'))
def when_click_word_press_hotkey(luteclient, word, hotkey):
    "Click word and press hotkey."
    luteclient.click_word(word)
    luteclient.press_hotkey(hotkey)


@when(parsers.parse('I hover over "{word}"'))
def when_hover(luteclient, word):
    "Hover over a term."
    els = luteclient.browser.find_by_text(word)
    assert len(els) == 1, f'have single "{word}"'
    els[0].mouse_over()


@when(parsers.parse('I press hotkey "{hotkey}"'))
def when_press_hotkey(luteclient, hotkey):
    "Click word and press hotkey."
    luteclient.press_hotkey(hotkey)


# Reading, paging


@when(parsers.parse("I click the footer green check"))
def when_click_footer_check(luteclient):
    "Click footer."
    luteclient.browser.find_by_id("footerMarkRestAsKnown").click()
    time.sleep(0.1)  # Leave this, remove and test fails.


@when(parsers.parse("I click the footer next page"))
def when_click_footer_next_page(luteclient):
    "Click footer."
    luteclient.browser.find_by_id("footerNextPage").click()
    time.sleep(0.1)  # Leave this, remove and test fails.
